# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import argparse
import errno
import logging
import re
import sys

import byov

from byoc import (
    errors as config_errors,
    registries,
)
from byov import (
    config,
    errors,
    options,
)
from byov.vms import (
    docker,
    scaleway,
    qemu,
    lxd,
)


logger = logging.getLogger(__name__)


# The available vm backends are registered here to ensure the config registry
# is loaded before we try to use it.
options.vm_class_registry.register('docker', docker.Docker,
                                   'docker container')
# FIXME: qemu really requires qemu-system-x86_64, qemu-img and genisoimage
# (respectively in packages qemu-system-x86, qemu-utils and genisoimage). This
# should be handled as a soft-dependency -- vila 2019-10-17
options.vm_class_registry.register('qemu', qemu.Qemu,
                                   'QEMU-based virtual machine')
options.vm_class_registry.register('lxd', lxd.Lxd,
                                   'Lxd virtual machine')
options.vm_class_registry.register('ephemeral-lxd', lxd.EphemeralLxd,
                                   'Linux container ephemeral virtual machine')
options.vm_class_registry.register('scaleway', scaleway.Scaleway,
                                   'Scaleway server')
if sys.version_info < (3,):
    try:
        from byov.vms import nova
    except ImportError:
        # No novaclient, no nova vms.
        pass
    else:
        # Only register nova if novaclient is available
        options.vm_class_registry.register(
            'nova', nova.NovaServer, 'Openstack Nova instance')

try:
    from byov.vms import ec2
except ImportError:
        # No boto3, no ec2 vms.
        pass
else:
    # Only register ec2 if boto3 is available
    options.vm_class_registry.register(
        'ec2', ec2.Ec2Server, 'Aws Ec2 instance')

# Import byov.conf.d/byov.py files so additional options (or other tweaks to
# byov.config) can be loaded.
if sys.version_info >= (3,):
    config.import_user_byovs()


class ArgParser(argparse.ArgumentParser):
    """A parser for the byovm script."""

    def __init__(self, name, description):
        script_name = 'byovm'
        super(ArgParser, self).__init__(
            prog='{} {}'.format(script_name, name),
            description=description)

    def parse_args(self, args=None, out=None, err=None):
        """Parse arguments, overridding stdout/stderr if provided.

        python argparse uses stdout and stderr for user convenience. For tests
        though, they need to be overriden so test can make assertion on their
        content.

        :params args: Defaults to sys.argv[1:].

        :param out: Defaults to sys.stdout.

        :param err: Defaults to sys.stderr.

        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(ArgParser, self).parse_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig

    def parse_known_args(self, args=None, out=None, err=None):
        """Parse known arguments, overridding stdout/stderr if provided.

        :params args: Defaults to sys.argv[1:].

        :param out: Defaults to sys.stdout.

        :param err: Defaults to sys.stderr.
        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(ArgParser, self).parse_known_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig


class CommandRegistry(registries.Registry):
    """A registry specialized for commands."""

    def register(self, cmd):
        super(CommandRegistry, self).register(
            cmd.name, cmd, help_string=cmd.description)


# All commands are registered here, defining what run() supports
command_registry = CommandRegistry()


class Command(object):

    name = 'Must be set by daughter classes'
    description = 'Must be set by daughter classes'

    def __init__(self, out=None, err=None, _vm=None):
        """Command constructor.

        :param out: A stream for command output.

        :param err: A stream for command errors.

        :param _vm: An existing vms.VM instance. For tests only.
        """
        self.vm = _vm
        if out is None:
            out = sys.stdout
        if err is None:
            err = sys.stderr
        self.out = out
        self.err = err
        self.parser = ArgParser(self.name, self.description)

    def parse_args(self, args):
        self.options = self.parser.parse_args(args, self.out, self.err)
        return self.options


class Help(Command):

    name = 'help'
    description = 'Describe byov configuration options.'

    def __init__(self, **kwargs):
        super(Help, self).__init__(**kwargs)
        # FIXME: Support regexps -- vila 2017-02-02
        self.parser.add_argument(
            'options', metavar='OPTION', nargs='*',
            help='Display help for each option'
            ' (Topics if none is given).')

    def run(self):
        if not self.options.options:
            self.out.write('Available Topics:\n')
            # Topics are only the ones known to byov.

            # FIXME: There should be a way to declare a topic for the user
            # -- vila 2018-11-04
            for topic in ['logging',
                          'vm', 'ec2', 'qemu', 'scaleway', 'nova', 'lxd',
                          'apt', 'ssh',
                          'launchpad', 'gitlab',
                          'debian', 'ubuntu']:
                names = [name for name, help
                         in options.option_registry.help.items()
                         if name.startswith(topic + '.')]
                self.out.write('\n{}\n'.format(topic))
                self.out.write('{}\n'.format('=' * len(topic)))
                for name in names:
                    option = options.option_registry.get(name)
                    self.out.write('{}: {}\n'.format(name, option.help))
            return
        # But all options are valid here, including the user-defined ones
        matched = set()
        for regexp in self.options.options:
            matcher = re.compile(regexp)
            for name in options.option_registry.keys():
                if name not in matched and matcher.search(name):
                    matched.add(name)
        if matched:
            for name in sorted(matched):
                option = options.option_registry.get(name)
                self.out.write('{}: {}\n'.format(name, option.help))
        else:
            msg = 'No options matched {}\n'.format(self.options.options)
            self.err.write(msg)
            return 1
        return 0


command_registry.register(Help)


class Version(Command):

    name = 'version'
    description = 'Output the byov version.'

    def run(self):
        self.out.write(byov.version() + '\n')
        return 0


command_registry.register(Version)


class VmCommand(Command):
    """A command applying to a given vm.

    This is not a command by itself but a base class for concrete commands.
    """

    def __init__(self, **kwargs):
        super(VmCommand, self).__init__(**kwargs)
        self.parser.add_argument(
            'vm_name',
            help='Virtual machine section in the configuration file.')
        self.parser.add_argument(
            '--option', '-O', metavar='OPTION=VALUE',
            action='append', dest='overrides', default=[],
            help='Override OPTION with provided VALUE. Can be repeated.')
        # override vm.name ?
        # Support @ containing lines of name=value and comments (aka a config
        # file including sections) or just -C <path>.conf ?

    def parse_args(self, args):
        super(VmCommand, self).parse_args(args)
        self.parse_vm_arg()
        return self.options

    def parse_vm_arg(self):
        self.vm_name = self.options.vm_name
        if self.vm is not None:
            # The vm is already populated
            return

        conf = config.VmStack(self.vm_name)
        # Append vm.name to overrides (which means it can't be overridden with
        # -O)
        self.options.overrides.append('vm.name={}'.format(self.vm_name))
        conf.cmdline_store.from_cmdline(self.options.overrides)
        try:
            kls_name = conf.get('vm.class', convert=False)
        except config_errors.OptionMandatoryValueError:
            # FIXME: In practice, this happens when the user makes a typo or
            # when running from the wrong place (where the vm is not defined in
            # any visible file) -- vila 2016-08-30
            raise errors.ByovError('"{name}" must define vm.class.',
                                   name=self.vm_name)
        # Even if defined, the value may be wrong
        try:
            kls = conf.get('vm.class')
        except config_errors.OptionMandatoryValueError:
            raise errors.InvalidVmClass(self.vm_name, kls_name)
        self.vm = kls(conf)
        # Now that we know the vm, we may configure logging 'logging' is not
        # configured until we reach this point, this means errors occuring
        # before this point can be logged without the user specified options.
        log_level = self.vm.conf.get('logging.level')
        log_format = self.vm.conf.get('logging.format')
        logging.basicConfig(level=log_level, format=log_format)
        return


class Config(VmCommand):

    name = 'config'
    description = 'Manage a virtual machine configuration.'

# FIXME: The full help should be the following but that doesn't fit well with
# the Help command for now -- vila 2014-03-26

# FIXME: Maybe a better way would be to delegate help handling to each command
# (instead of a single help command). From there, 'help config' could accept an
# option name or regexp and display the help for the matching options.
# -- vila 2013-08-30

# Display the active value for option NAME.
#
# If NAME is not given, --all .* is implied (all options are displayed).
#
# If --all is specified, NAME is interpreted as a regular expression and all
# matching options are displayed (without resolving option references in the
# value). The active value is the first one displayed for each option.
#
# NAME=value without spaces sets the value.
#
# --remove NAME remove the option definition.

    def __init__(self, **kwargs):
        super(Config, self).__init__(**kwargs)
        self.parser.add_argument(
            'name', nargs='?', help='''The option name.''')
        self.parser.add_argument(
            '--remove', '-r', action="store_true", help='Remove the option.'),
        self.parser.add_argument(
            '--all', '-a', action="store_true",
            help='Display all the defined values for the matching options.'),

    def parse_args(self, args):
        # Work around argparse not providing a proper way to mix positional
        # arguments and options. If an option arg is used, the 'name'
        # positional arg is not set and ends up in 'rest'.
        self.options, rest = self.parser.parse_known_args(
            args, self.out, self.err)
        self.parse_vm_arg()
        o = self.options
        if o.all or o.remove:
            # Get 'name' back from 'rest' if available
            if rest:
                o.name = rest.pop(0)
        # If we still have args left, it's an error
        if rest:
            err_orig = sys.stderr
            try:
                sys.stderr = self.err
                msg = 'unrecognized arguments: {}'
                self.parser.error(msg.format(' '.join(rest)))
            finally:
                sys.stderr = err_orig
        # Raise errors on option incompatibilities.
        if o.remove:
            if o.all:
                raise errors.ByovError(
                    '--remove and --all are mutually exclusive.')
            if o.name is None:
                raise errors.ByovError(
                    '--remove expects an option to remove.')
        if o.name is not None and '=' in o.name:
            if o.all:
                raise errors.ByovError('Only one option can be set.')
        return o

    def list(self, name):
        # Display the value as a string
        value = self.vm.conf.get(name, expand=True, convert=False)
        if value is not None:
            # No quoting needed (for now)
            self.out.write(value)
        else:
            raise errors.ConfigOptionNotFound(name)

    def list_matching(self, re_str):
        not_a_section_id = object()
        cur_store_id = None
        cur_section_id = not_a_section_id
        names = re.compile(re_str)
        for (store, section, name, value) in self.vm.conf.iter_options():
            if names.search(name) is None:
                # not matching option, ignore
                continue
            if cur_store_id != store.id:
                # Explain where the options are defined
                self.out.write('{}:\n'.format(store.id,))
                cur_store_id = store.id
                cur_section_id = not_a_section_id
            if (cur_section_id != section.id):
                if section.id is None:
                    section_displayed = ''
                else:
                    section_displayed = section.id
                # Display the section id as it appears in the store
                self.out.write('  [{}]\n'.format(section_displayed))
            cur_section_id = section.id
            # No quoting needed (for now)
            self.out.write('  {} = {}\n'.format(name, value))

    def run(self):
        o = self.options
        if o.remove:
            try:
                self.vm.conf.remove(o.name)
            except config_errors.NoSuchConfigOption as e:
                raise errors.ConfigOptionNotFound(e.name)
            self.vm.conf.store.save_changes()
        elif o.name is not None and '=' in o.name:
            # Set the option value
            name, value = o.name.split('=', 1)
            self.vm.conf.set(name, value)
            self.vm.conf.store.save_changes()
        else:
            # List the options
            if o.name is None:
                # Defaults to all options
                o.name = '.*'
                o.all = True
            if o.all:
                self.list_matching(o.name)
            elif o.name.startswith('@'):
                try:
                    self.out.write(self.vm.expand_file(o.name[1:]))
                except IOError as e:
                    # python2 does not provide FileNotFoundError
                    if e.errno == errno.ENOENT:
                        raise errors.ConfigPathNotFound(o.name[1:], o.name)
                    else:
                        raise
            else:
                self.list(o.name)
        return 0


command_registry.register(Config)


class Setup(VmCommand):

    name = 'setup'
    description = 'Setup a virtual machine.'

    def __init__(self, **kwargs):
        super(Setup, self).__init__(**kwargs)
        self.parser.add_argument(
            '--download', '-d', action="store_true",
            help='Force download of the required image.')
        self.parser.add_argument(
            '--ssh-keygen', '-k', action="store_true",
            help="Generate the ssh keys defined in 'ssh.server_keys'.")
        self.parser.add_argument(
            '--force', '-f', action="store_true",
            help='Force setup even if the vm is running.')

    def run(self):
        state = self.vm.state()
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state in ('running', 'RUNNING'):
            if self.options.force:
                self.vm.teardown(force=self.options.force)
                state = self.vm.state()
            else:
                raise errors.VmRunning(self.vm_name)
        elif state in('shut off', 'STOPPED'):
            self.vm.teardown()
        self.vm.ssh_keygen(force=self.options.ssh_keygen)

        self.vm.setup()
        if self.vm.conf.get('vm.poweroff'):
            self.vm.stop()
        return 0


command_registry.register(Setup)


class Digest(VmCommand):

    name = 'digest'
    description = 'Display the digest of a virtual machine.'

    def parse_args(self, args):
        # FIXME: Another case where 'must define vm.class' sucks
        # -- vila 2016-11-14
        try:
            self.options = super(Digest, self).parse_args(args)
        except errors.ByovError as e:
            if 'must define vm.class' in e.fmt:
                self.options = None
        return self.options

    def run(self):
        if self.options is None:
            digest = 'UNKNOWN'
        else:
            digest = self.vm.hash_setup()
        if sys.version_info < (3,):
            digest = digest.decode('utf8')
        self.out.write(digest)
        return 0


command_registry.register(Digest)


class Status(VmCommand):

    name = 'status'
    description = 'Display the status of a virtual machine.'

    def run(self):
        state = self.vm.state()
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state in ('shut off', 'STOPPED'):
            state = 'STOPPED'
        elif state in ('running', 'RUNNING'):
            state = 'RUNNING'
        self.out.write(state + '\n')
        return 0


command_registry.register(Status)


class Start(VmCommand):

    name = 'start'
    description = 'Start an existing virtual machine.'

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state in ('running', 'RUNNING'):
            raise errors.VmRunning(self.vm_name)
        self.vm.start()
        return 0


command_registry.register(Start)


class Shell(VmCommand):

    name = 'shell'
    description = (
        'Start a shell, run a command or a script inside a virtual machine.')

    def __init__(self, **kwargs):
        super(Shell, self).__init__(**kwargs)
        self.parser.add_argument(
            'command', help='The command to run inside the vm.'
            ' Use a @ prefix to use a local script instead', nargs='?')
        self.parser.add_argument(
            'args', help='The arguments for the command to run on the vm.',
            nargs=argparse.REMAINDER)

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state not in ('running', 'RUNNING'):
            raise errors.VmNotRunning(self.vm_name)
        ret, out, err = self.vm.shell(self.options.command, *self.options.args)
        # out and err are displayed separately rather than interspred as in
        # real life... but hey, is it that bad to lose the hard-to-figure-out
        # chaos ?
        print(out, file=self.out, end='')
        print(err, file=self.err, end='')
        return ret


command_registry.register(Shell)


class SshRegister(VmCommand):

    name = 'ssh-register'
    description = 'Register a host in the vm for ssh access.'

    def __init__(self, **kwargs):
        super(SshRegister, self).__init__(**kwargs)
        # FIXME: Support using an existing key ? -- vila 2018-10-21
        self.parser.add_argument(
            '--type', '-t', default='rsa',
            help='The ssh key type.')
        self.parser.add_argument(
            'host', help='The host to register.')

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state not in ('running', 'RUNNING'):
            raise errors.VmNotRunning(self.vm_name)
        # FIXME: Duplicated from VmCommand.parse_vm_arg -- vila 2018-10-22
        host_conf = config.VmStack(self.options.host)
        try:
            kls = host_conf.get('vm.class')
        except config_errors.OptionMandatoryValueError:
            raise errors.VmUnknown(self.options.host)
        self.host = kls(host_conf)
        host_state = self.host.state()
        if host_state is None:
            raise errors.VmUnknown(self.options.host)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if host_state not in ('running', 'RUNNING'):
            raise errors.VmNotRunning(self.options.host)
        # Get the host key from the host itself (hard to MITM ;-)
        keyscan_cmd = ['ssh-keyscan', '-t', self.options.type, 'localhost']
        ret, out, err = self.host.shell_captured(*keyscan_cmd)
        if ret:
            raise errors.CommandError(keyscan_cmd, ret, out, err)
        # Put the key in the vm known_hosts
        host_name = host_conf.get('vm.name')
        host_ip = host_conf.get('vm.ip')
        host_def = ('"' + out.replace('localhost ',
                                      '{},{} '.format(host_name, host_ip)) +
                    '"')
        knownhosts_cmd = ['echo', '-n', host_def, '>>', '~/.ssh/known_hosts']
        ret, out, err = self.vm.shell_captured(*knownhosts_cmd)
        if ret:
            raise errors.CommandError(knownhosts_cmd, ret, out, err)
        return ret


command_registry.register(SshRegister)


class SshAuthorize(VmCommand):

    name = 'ssh-authorize'
    description = 'Authorize a key in the vm for ssh access.'

    def __init__(self, **kwargs):
        super(SshAuthorize, self).__init__(**kwargs)
        self.parser.add_argument(
            'key', help='The path to the public key to add.')

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state not in ('running', 'RUNNING'):
            raise errors.VmNotRunning(self.vm_name)
        with open(self.options.key) as f:
            key = f.read()
        authorize_cmd = ['cat', '-', '>>', '~/.ssh/authorized_keys']
        ret, out, err = self.vm.shell_captured(*authorize_cmd, cmd_input=key)
        if ret:
            raise errors.CommandError(authorize_cmd, ret, out, err)
        return ret


command_registry.register(SshAuthorize)


class Pull(VmCommand):

    name = 'pull'
    description = 'Pull a remote file from an existing virtual machine.'

    def __init__(self, **kwargs):
        super(Pull, self).__init__(**kwargs)
        self.parser.add_argument(
            'remote', metavar='REMOTE',
            help='The remote file in the vm to pull from.')
        self.parser.add_argument(
            'local', metavar='LOCAL',
            help='The local file.')

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state not in ('running', 'RUNNING'):
            raise errors.VmNotRunning(self.vm_name)
        retcode = self.vm.pull(self.options.remote, self.options.local)
        return retcode


command_registry.register(Pull)


class Push(VmCommand):

    name = 'push'
    description = 'Push local file to an existing virtual machine.'

    def __init__(self, **kwargs):
        super(Push, self).__init__(**kwargs)
        self.parser.add_argument(
            'local', metavar='LOCAL',
            help='The local file to push. Option expanded if prefixed with @.')
        self.parser.add_argument(
            'remote', metavar='REMOTE',
            help='The remote file in the vm to push to.')

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state not in ('running', 'RUNNING'):
            raise errors.VmNotRunning(self.vm_name)
        retcode = self.vm.push(self.options.local, self.options.remote)
        return retcode


command_registry.register(Push)


class Stop(VmCommand):

    name = 'stop'
    description = 'Stop an existing virtual machine.'

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state not in ('running', 'RUNNING'):
            raise errors.VmNotRunning(self.vm_name)
        self.vm.stop()
        return 0


command_registry.register(Stop)


class Publish(VmCommand):

    name = 'publish'
    description = 'Publish a virtual machine image.'

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state in ('running', 'RUNNING'):
            raise errors.VmRunning(self.vm_name)
        self.vm.publish()
        return 0


command_registry.register(Publish)


class Teardown(VmCommand):

    name = 'teardown'
    description = 'Teardown a virtual machine.'

    def __init__(self, **kwargs):
        super(Teardown, self).__init__(**kwargs)
        self.parser.add_argument(
            '--force', '-f', action="store_true",
            help='Force teardown even if the vm is running.')

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        if state in ('running', 'RUNNING') and not self.options.force:
            raise errors.VmRunning(self.vm_name)
        self.vm.teardown(force=self.options.force)
        return 0


command_registry.register(Teardown)

# Now that we have all the commands registered, we can create a place holder
# option for help.
commands_help = '\n' + ''.join([
    '\t{}: {}\n'.format(cmd_name,
                        command_registry.get(cmd_name).description)
    for cmd_name in command_registry.keys()])
options.register(options.Option('byov.commands', default=None,
                                help_string=commands_help))


def run(args=None, out=None, err=None):
    if args is None:
        args = sys.argv[1:]
    if not args:
        cmd = Help()
        cmd_name = 'help'
    else:
        cmd_name = args[0]
        args = args[1:]
        try:
            cmd_class = command_registry.get(cmd_name)
            cmd = cmd_class(out=out, err=err)
        except KeyError:
            cmd = Help(out=out, err=err)
            args = [cmd_name]
    try:
        cmd.parse_args(args)
        return cmd.run()
    except Exception as e:
        logger.debug('{} failed'.format(cmd_name), exc_info=True)
        logger.error('{} failed: {!r}'.format(cmd_name, e))
        return -1



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import io
import errno
import os
import sys
import unittest

from byoc import errors as conf_errors
from byov import (
    config,
    errors,
)
from byov.tests import (
    features,
    fixtures,
)
from byov.vms import lxd


@features.requires(features.tests_config)
class TestWithEphemeralLxd(unittest.TestCase):

    def setUp(self):
        super(TestWithEphemeralLxd, self).setUp()
        fixtures.setup_tests_config(self)
        features.requires_existing_vm(self, 'byov-tests-lxd-ubuntu')
        fixtures.set_uniq_vm_name(self)
        fixtures.override_logging(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = ephemeral-lxd
vm.backing = byov-tests-lxd-ubuntu
'''.format(vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()
        self.vm = lxd.EphemeralLxd(config.VmStack(self.vm_name))
        self.addCleanup(self.vm.stop)
        self.vm.start()


# FIXME: parametrized by distribution -- vila 2017-12-06
class TestUpload(TestWithEphemeralLxd):

    def test_binary_file(self):
        self.vm.upload('/usr/bin/yes', '/home/ubuntu/yes')
        self.assertEqual(
            (0, '', ''),
            self.vm.shell_captured('test', '-f', '/home/ubuntu/yes'))

    def test_unknown_file(self):
        # python2 doesn't have FileNotFoundError
        with self.assertRaises(OSError) as cm:
            self.vm.upload('I-dont-exist', 'not-used')
        self.assertEqual(errno.ENOENT, cm.exception.errno)

    def test_unknown_remote_dir(self):
        # This is mainly testing the error handling for the receiving pipe
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.upload('/usr/bin/yes', '/I-dont-exist/yes')
        self.assertEquals(1, cm.exception.retcode)
        self.assertTrue('/I-dont-exist/yes' in cm.exception.err)

    def test_empty_file(self):
        with open('empty', 'w') as f:
            f.write('')
        self.vm.upload('empty',
                       '/home/{}/empty'.format(self.vm.conf.get('vm.user')))
        self.assertEqual(
            (0, '', ''),
            self.vm.shell_captured(
                'test', '-f',
                '/home/{}/empty'.format(self.vm.conf.get('vm.user'))))


class TestRunScript(TestWithEphemeralLxd):

    # FIXME: Popen(stdout=None) can be redirected so we can't capture it which
    # means this test leaks, disabling until blackbox tests are defined (using
    # the byovm script). -- vila 2016-07-04
    def xtest_simple_script(self):
        with open('script', 'w')as f:
            f.write('#!/bin/sh\nwhoami\n')
        os.chmod('script', 0o755)
        fixtures.patch(self, sys, 'stdout', io.StringIO())
        self.vm.run_script('script', captured=False)

    def test_script_fails_with_params(self):
        with open('script', 'w')as f:
            f.write('#!/bin/sh\necho $*\nfalse\n')
        os.chmod('script', 0o755)
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.run_script('script', ['foo'])
        self.assertEqual(1, cm.exception.retcode)
        self.assertEqual('foo\n', cm.exception.out)
        self.assertEqual('', cm.exception.err)


class TestSetupScripts(TestWithEphemeralLxd):
    """Test vm.setup_scripts on a started vm as that's faster."""

    def test_no_scripts(self):
        # The following shouldn't fail
        self.vm.run_setup_scripts()

    def test_unknown_script(self):
        self.vm.conf.set('vm.setup_scripts', 'I-dont-exist')
        with self.assertRaises(OSError) as cm:
            self.vm.run_setup_scripts()
        self.assertEquals(errno.ENOENT, cm.exception.errno)
        self.assertTrue('I-dont-exist' in str(cm.exception))

    def test_several_scripts(self):
        with open('script1', 'w')as f:
            f.write('#!/bin/sh\ntouch script1-ran\n')
        os.chmod('script1', 0o755)
        with open('script2', 'w')as f:
            f.write('#!/bin/sh\ntouch script2-ran\n')
        os.chmod('script2', 0o755)
        self.vm.conf.set('vm.setup_scripts', 'script1, script2')
        self.vm.run_setup_scripts()
        ret, out, err = self.vm.shell_captured('ls', 'script*ran')
        self.assertEqual(0, ret)
        self.assertEqual('script1-ran\nscript2-ran\n', out)
        self.assertEqual('', err)

    def test_expand_options_in_script(self):
        script_name = 'script1'
        self.vm.conf.set('script_name', script_name)
        with open(script_name, 'w')as f:
            f.write('#!/bin/sh\ntouch {script_name}-ran\n')
        os.chmod(script_name, 0o755)
        self.vm.conf.set('vm.setup_scripts', script_name)
        self.vm.run_setup_scripts()
        ret, out, err = self.vm.shell_captured(
            'ls', '{}*ran'.format(script_name))
        self.assertEqual(0, ret)
        self.assertEqual('{}-ran\n'.format(script_name), out)
        self.assertEqual('', err)

    def test_expansion_error(self):
        with open('script1', 'w')as f:
            f.write('#!/bin/sh\ntouch {I_dont_exist}-ran\n')
        os.chmod('script1', 0o755)
        self.vm.conf.set('vm.setup_scripts', 'script1')
        with self.assertRaises(conf_errors.ExpandingUnknownOption) as cm:
            self.vm.run_setup_scripts()
        self.assertEqual('I_dont_exist', cm.exception.name)

    def test_home_script(self):
        script_path = '~/script'
        expanded_script_path = os.path.expanduser(script_path)
        with open(expanded_script_path, 'w')as f:
            f.write('#!/bin/sh\ntouch script-ran\n')
        os.chmod(expanded_script_path, 0o755)
        self.vm.conf.set('vm.setup_scripts', script_path)
        self.vm.run_setup_scripts()
        ret, out, err = self.vm.shell_captured('ls', 'script*ran')
        self.assertEqual(0, ret)
        self.assertEqual('script-ran\n', out)
        self.assertEqual('', err)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2021 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import logging
import re
import time

import boto3
import botocore

from byov import (
    errors,
    options,
    vms,
)

logger = logging.getLogger(__name__)

# FIXME: There is a x86_64_mac arch but nothing for m1 -- vila 2021-11-29

# aws uses a different nomenclature. Probably closer to linux than ubuntu
# or lxd.
ec2_architectures = dict(
    i386='i386',
    amd64='x86_64',
    arm64='arm64',
)
# Which owners are responsible for producing AMIs
# This is used to filter the huge number of available images.
# 'self' refers to the user creating images
ec2_image_owners = dict(
    # Canonical https://cloud-images.ubuntu.com/locator/ec2/
    ubuntu=['self', '099720109477'],
    # debian-cloud https://wiki.debian.org/Cloud/AmazonEC2Image/
    debian=['self', '136693071363'],
    # alias for several owners
    amazon=['self', 'amazon'],
)


def get_ec2_client(conf):
    session = boto3.session.Session()
    ec2_client = session.client(
        "ec2",
        aws_access_key_id=conf.get('aws.key'),
        aws_secret_access_key=conf.get('aws.secret'),
        aws_session_token=conf.get('aws.token'),
        region_name=conf.get('aws.region'))
    return ec2_client


class Ec2ServerException(errors.ByovError):
    pass


class Ec2Client(object):
    """An ec2 client re-trying requests on known transient failures."""

    def __init__(self, conf, **kwargs):
        self.ec2 = get_ec2_client(conf)

    def retry(self, func, *args, **kwargs):
        # No known transient failures at this point so no retry
        try:
            return func(*args, **kwargs)
        except botocore.exceptions.ClientError as e:
            logger.error('{} failed'.format(func.__name__), exc_info=True)
            raise Ec2ServerException('{name} failed: {boto_exc}',
                                     name=func.__name__, boto_exc=e)

    def instance(self, instance_id):
        res = self.retry(self.ec2.describe_instances,
                         InstanceIds=[instance_id])
        return res['Reservations'][0]['Instances'][0]

    def instance_types(self, types, free_tier):
        func_name = self.instance_types.__name__
        valid_types = []
        pager = self.retry(self.ec2.get_paginator, 'describe_instance_types')
        more = True
        # MISSINGTEST: the NextToken trick requires reducing the number of
        # types response to trigger more calls -- vila 1021-11-26

        # Be prepared to receive a StartingToken.
        pconf = dict()
        while more:
            try:
                for res in self.retry(pager.paginate, InstanceTypes=types,
                                      PaginationConfig=pconf):
                    rtypes = res['InstanceTypes']
                    if not rtypes:
                        more = False
                        break
                    for t in rtypes:
                        if t['FreeTierEligible'] != free_tier:
                            continue
                        valid_types.append(t['InstanceType'])
            except botocore.exceptions.ClientError as e:
                logger.error('{} failed'.format(func_name), exc_info=True)
                raise Ec2ServerException('boto failed: {boto_exc}',
                                         name=func_name, boto_exc=e)
            next_token = res.get('NextToken', None)
            if next_token is None:
                more = False
            else:
                # If there is a NextToken in the last received response, go one
                # more round
                pconf['StartingToken'] = next_token
        return valid_types

    def images(self, arch, owners, image_id=None):
        valid_images = []
        filters = []
        # Reduce the number of images returned by filtering on some known
        # constraints
        filters.append(dict(Name='state', Values=['available']))
        filters.append(dict(Name='virtualization-type', Values=['hvm']))
        filters.append(dict(Name='architecture', Values=[arch]))
        filters.append(dict(Name='image-type', Values=['machine']))
        kwargs = dict(Filters=filters)
        # amazon (an alias for several owners) produces Amazon Linux images
        # Canonical (099720109477) produces ubuntu images
        # RedHat (309956199498) produces RedHat images
        # debian (136693071363) produces Debian 11 (bullseye, stable) and
        # Debian 10 (buster, oldstable)
        # https://wiki.debian.org/Cloud/AmazonEC2Image/

        kwargs['Owners'] = owners
        if image_id is not None:
            kwargs['ImageIds'] = [image_id]

        res = self.retry(self.ec2.describe_images, **kwargs)
        rimgs = res['Images']
        for i in rimgs:
            valid_images.append(i)
        return valid_images

    def create_server(self, name, itype, image_id, user_data, sec_groups,
                      subnet, tags):
        kwargs = dict(MinCount=1, MaxCount=1,
                      TagSpecifications=tags,
                      InstanceType=itype, ImageId=image_id,
                      SecurityGroupIds=sec_groups,
                      SubnetId=subnet,
                      UserData=user_data)
        res = self.retry(self.ec2.run_instances, **kwargs)
        if res['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise Ec2ServerException('{name} creation failed: {res}',
                                     name=name, res=res)
        return res['Instances'][0]['InstanceId']

    def terminate_server(self, instance_id):
        res = self.retry(self.ec2.terminate_instances,
                         InstanceIds=[instance_id])
        if res['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise Ec2ServerException('{id} terminate failed : {res}',
                                     id=instance_id, res=res)

    def start_server(self, instance_id):
        res = self.retry(self.ec2.start_instances,
                         InstanceIds=[instance_id])
        if res['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise Ec2ServerException('{id} start failed : {res}',
                                     id=instance_id, res=res)

    def stop_server(self, instance_id):
        res = self.retry(self.ec2.stop_instances,
                         InstanceIds=[instance_id])
        if res['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise Ec2ServerException('{id} stop failed : {res}',
                                     id=instance_id, res=res)

    def create_image(self, instance_id, image_name, tags):
        res = self.retry(
            self.ec2.create_image, InstanceId=instance_id, Name=image_name,
            TagSpecifications=tags)
        image_id = res['ImageId']
        return image_id

    def deregister_image(self, image_id):
        self.retry(
            self.ec2.deregister_image, ImageId=image_id)

    def delete_snapshot(self, snapshot_id):
        self.retry(self.ec2.delete_snapshot, SnapshotId=snapshot_id)


class Ec2Server(vms.VM):

    ec2_client_class = Ec2Client
    setup_ip_timeouts = 'ec2.setup_ip.timeouts'
    setup_ssh_timeouts = 'ec2.setup_ssh.timeouts'
    existing_vm_options = vms.VM.existing_vm_options + [
        'ec2.image',
        'ec2.image.id',
        'ec2.instance.id',
        'ec2.security.groups',
        'ec2.subnet',
    ]

    def __init__(self, conf):
        super().__init__(conf)
        self.ec2 = self.build_ec2_client()
        self.econf.set('vm.final_message', 'testbed setup completed.')

    def build_ec2_client(self):
        ec2_client = self.ec2_client_class(self.conf)
        return ec2_client

    def state(self):
        instance_id = self.conf.get('ec2.instance.id')
        if instance_id is None:
            return 'UNKNOWN'

        try:
            instance = self.ec2.instance(instance_id)
        except Ec2ServerException:
            logger.error('{} raised'.format(instance_id), exc_info=True)
            return 'UNKNOWN'
        # The instance may remain in the terminated state for some
        # time. stopping and shutting down are transient states but the vm is
        # still running.
        ec2_states = dict(pending='STARTING',
                          running='RUNNING',
                          stopping='STOPPING',
                          stopped='STOPPED',
                          terminated='UNKNOWN')
        ec2_states['shutting-down'] = 'RUNNING'
        return ec2_states[instance['State']['Name']]

    def find_instance_type(self):
        free = self.conf.get('ec2.free_tier')
        wanted_types = self.conf.get('ec2.instance.types')
        if not wanted_types:
            raise Ec2ServerException('ec2.instance.types must be set')
        logger.debug('Searching for a valid instance type in {}'.format(
            ' '.join(wanted_types)))
        existing_types = self.ec2.instance_types(wanted_types, free)
        for t in wanted_types:
            for existing in existing_types:
                if t == existing:
                    return existing
        raise Ec2ServerException(
            'None of [{types}] can be found', types=','.join(wanted_types))

    def find_ec2_image(self, image_id=None):
        """Find the most recent image matching a given name.

        If image_id is set and exists take it even if it doesn't match
        {ec2.image}.
        """
        if image_id is None:
            # Default to the one set by the user if any
            image_id = self.conf.get('ec2.image.id')
        # We want to select an image either by its name, its AMI id or from a
        # combination distribution/release/arch. If several matches are found,
        # the most recent is taken.
        ec2_image_name = self.conf.get('ec2.image')
        if ec2_image_name:
            image_re = re.compile(ec2_image_name)
        else:
            # Use the distribution default
            distro = self.conf.get('vm.distribution')
            regexp = options.ec2_image_re_registry.get(distro)
            image_re = re.compile(self.conf.expand_options(regexp))

        arch = ec2_architectures[self.conf.get('vm.architecture')]
        logger.debug('Searching for image {} {}...'.format(image_re, image_id))
        owners = ec2_image_owners.get(self.conf.get('vm.distribution'),
                                      ['self'])
        existing_images = self.ec2.images(arch, owners, image_id=image_id)
        logger.debug("{} existing images".format(len(existing_images)))
        candidates = []
        for existing in existing_images:
            if image_id and image_id == existing['ImageId']:
                # Shortcut name based selection, user knows better.
                return existing
            m = image_re.search(existing['Name'])
            if m is not None:
                candidates.append(existing)
        if candidates:
            logger.debug("{} candidate images".format(len(candidates)))
            # Use the most recent one
            candidates.sort(key=lambda c: c['CreationDate'], reverse=True)
            return candidates[0]
        if image_id:
            raise Ec2ServerException(
                'Image "{id}" cannot be found', id=image_id)
        else:
            raise Ec2ServerException(
                'No matches for "{name}" can be found', name=image_re)

    def make_tag_specs(self, res_type, *args):
        it = iter(args)
        tag_spec = dict(
            ResourceType=res_type,
            Tags=[dict(Key=k, Value=next(it)) for k in it])
        return tag_spec

    def instance_tags(self):
        # instance name is implemented as a tag: deal with it.
        tag_spec = self.make_tag_specs(
            'instance', 'Name', self.conf.get('vm.name'),
            *self.conf.get('ec2.instance.tags'))
        return [tag_spec]

    def image_tags(self):
        image_tag_spec = self.make_tag_specs(
            'image', 'Name', self.conf.get('vm.name'),
            *self.conf.get('ec2.image.tags'))
        # Also tag the snapshots
        snapshot_tag_spec = self.make_tag_specs(
            'snapshot', 'Name', self.conf.get('vm.name'),
            *self.conf.get('ec2.image.tags'))
        return [image_tag_spec, snapshot_tag_spec]

    # FIXME: This should save the console whether or not the setup fails
    # -- vila 2021-11-30
    def setup(self):
        logger.info('Setting up {} ec2 server...'.format(
            self.conf.get('vm.name')))
        itype = self.find_instance_type()
        image_id = self.find_ec2_image()['ImageId']
        self.create_user_data()
        with open(self._user_data_path) as f:
            user_data = f.read()
        logger.debug('Creating {} ec2 server...'.format(
            self.conf.get('vm.name')))
        instance_id = self.ec2.create_server(
            name=self.conf.get('vm.name'), itype=itype, image_id=image_id,
            user_data=user_data,
            sec_groups=self.conf.get('ec2.security.groups'),
            subnet=self.conf.get('ec2.subnet'),
            tags=self.instance_tags())
        self.econf.set('ec2.instance.id', instance_id)
        self.econf.store.save_changes()  # Recovery checkpoint
        self.wait_for_running_instance()
        self.wait_for_ip()
        self.wait_for_ssh()
        self.wait_for_cloud_init()
        self.setup_over_ssh()

    def hash_setup(self):
        hasher = hashlib.md5()
        # Seed with the digest from the base class to capture common options
        hasher.update(super(Ec2Server, self).hash_setup().encode('utf8'))
        options = self.conf.get('ec2.setup.digest.options')
        if options:
            for opt in options:
                value = self.conf.get(opt, convert=False)
                self.hash_value(hasher, value)
        return hasher.hexdigest()

    def wait_for_state(self, target_states, timeouts_name):
        me = self.wait_for_state.__name__
        vm_name = self.conf.get('vm.name')
        tmouts = self.conf.get(timeouts_name)
        logger.debug('{} {} to reach {} until {}...'.format(
            me, vm_name, target_states, tmouts))
        state = 'UNKNOWN'
        for attempt, sleep in enumerate(tmouts, start=1):
            if attempt > 1:
                logger.debug('Re-trying {} {}/{}'.format(
                    me, attempt, tmouts.retries))
            state = self.state()
            logger.debug('{} is in {} state...'.format(vm_name, state))
            if state in target_states:
                break
            logger.debug('Sleeping {:.2f} seconds for {} {}/{}'.format(
                sleep, me, attempt, tmouts.retries))
            time.sleep(sleep)
        if state not in target_states:
            instance_id = self.conf.get('ec2.instance.id')
            raise Ec2ServerException(
                '{} never reached {} (was: {} last)'.format(
                    instance_id, target_states, state))

    def wait_for_running_instance(self):
        self.wait_for_state(['RUNNING'], 'ec2.boot.timeouts')

    def discover_ip(self):
        instance = self.ec2.instance(self.conf.get('ec2.instance.id'))
        # Just use the public IP for now
        return instance['PublicIpAddress']

    def wait_for_cloud_init(self):
        self._wait_for_cloud_init('ec2.cloud_init.timeouts')

    def wait_for_image_state(self, image_id, target_states, timeouts_name):
        me = self.wait_for_state.__name__
        tmouts = self.conf.get(timeouts_name)
        logger.debug('{} {} to reach {} until {}...'.format(
            me, image_id, target_states, tmouts))
        state = 'unknown'
        for attempt, sleep in enumerate(tmouts, start=1):
            if attempt > 1:
                logger.debug('Re-trying {} {}/{}'.format(
                    me, attempt, tmouts.retries))
                try:
                    image = self.find_ec2_image(image_id)
                    state = image['State']
                    logger.debug('{} is in {} state...'.format(
                        image_id, state))
                    if state in target_states:
                        break
                except Ec2ServerException:
                    logger.error('{} does not exist yet'.format(image_id),
                                 exc_info=True)
            logger.debug('Sleeping {:.2f} seconds for {} {}/{}'.format(
                sleep, me, attempt, tmouts.retries))
            time.sleep(sleep)
        if state not in target_states:
            raise Ec2ServerException(
                '{} never reached {} (was: {} last)'.format(
                    image_id, target_states, state))

    def start(self):
        super(Ec2Server, self).start()
        logger.info('Starting {} ec2 server...'.format(
            self.conf.get('vm.name')))
        self.ec2.start_server(self.conf.get('ec2.instance.id'))
        self.wait_for_running_instance()
        self.wait_for_ip()
        self.wait_for_ssh()
        self.save_existing_config()

    def stop(self):
        logger.info('Stopping {} ec2 server...'.format(
            self.conf.get('vm.name')))
        self.ec2.stop_server(self.conf.get('ec2.instance.id'))
        self.wait_for_state(['STOPPED'], 'ec2.stop.timeouts')

    def teardown(self, force=False):
        logger.info('Tearing down {} ec2 server...'.format(
            self.conf.get('vm.name')))
        if force and self.state() == 'RUNNING':
            self.stop()
        instance_id = self.conf.get('ec2.instance.id')
        if instance_id is not None:
            logger.info('Deleting instance {}'.format(instance_id))
            self.ec2.terminate_server(instance_id)
            self.wait_for_state(['TERMINATED', 'UNKNOWN'],
                                'ec2.terminate.timeouts')
        super(Ec2Server, self).teardown()

    def publish(self):
        image_name = self.conf.get('vm.published_as')
        logger.info('Publishing ec2 image {} from {}...'.format(
            image_name, self.conf.get('vm.name')))
        image_id = self.ec2.create_image(self.conf.get('ec2.instance.id'),
                                         image_name,
                                         tags=self.image_tags())
        self.wait_for_image_state(image_id, ['available'],
                                  'ec2.create_image.timeouts')
        self.econf.set('ec2.published.id', image_id)
        self.econf.store.save_changes()

    def unpublish(self):
        image_name = self.conf.get('vm.published_as')
        image_id = self.conf.get('ec2.published.id')
        if image_id is None:
            return
        logger.info('Un-publishing ec2 image {} {} from {}...'.format(
            image_name, image_id, self.conf.get('vm.name')))
        image = self.find_ec2_image(image_id)
        snapshot_ids = []
        for mappings in image['BlockDeviceMappings']:
            snapshot_id = mappings.get('Ebs', {}).get('SnapshotId', None)
            if snapshot_id is not None:
                snapshot_ids.append(snapshot_id)
        self.ec2.deregister_image(image_id)
        # snaphots used by the image must be deleted after the image is
        # deregistered.
        for snapshot_id in snapshot_ids:
            logger.debug('Deleting snapshot {}'.format(snapshot_id))
            self.ec2.delete_snapshot(snapshot_id)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import os
import unittest


from byov import (
    config,
    errors,
    vms,
    ssh,
)
from byov.tests import (
    features,
    fixtures,
)


class TestMetaData(unittest.TestCase):

    def setUp(self):
        super(TestMetaData, self).setUp()
        fixtures.isolate_from_disk(self)
        self.conf = config.VmStack('foo')
        self.vm = vms.VM(self.conf)
        images_dir = os.path.join(self.uniq_dir, 'images')
        self.conf.store._load_from_string('''
vm.name=foo
vm.images_dir={images_dir}
'''.format(images_dir=images_dir))

    def test_create_meta_data(self):
        self.vm.create_meta_data()
        self.assertTrue(os.path.exists(self.vm.config_dir_path()))
        self.assertTrue(os.path.exists(self.vm._meta_data_path))
        with open(self.vm._meta_data_path) as f:
            meta_data = f.readlines()
        self.assertEqual(2, len(meta_data))
        self.assertEqual('instance-id: foo\n', meta_data[0])
        self.assertEqual('local-hostname: foo\n', meta_data[1])


class TestCreateUserData(unittest.TestCase):

    def setUp(self):
        super(TestCreateUserData, self).setUp()
        fixtures.setup_tests_config(self)
        self.conf = config.VmStack('foo')
        self.vm = vms.VM(self.conf)

    def test_empty_config(self):
        self.conf.set('vm.name', 'foo')
        self.conf.set('vm.class', 'lxd')
        self.vm.create_user_data()
        self.assertTrue(os.path.exists(self.vm.config_dir_path()))
        self.assertTrue(os.path.exists(self.vm._user_data_path))
        with open(self.vm._user_data_path) as f:
            user_data = f.readlines()
        # We care about the two first lines only here, checking the format (or
        # cloud-init is confused)
        self.assertEqual('#cloud-config-archive\n', user_data[0])
        # yaml seems to flap between different formats changing ' to " and \n
        # to \\n depending on the content length...
        self.assertTrue(user_data[1].startswith('- {content: '))
        self.assertTrue('#cloud-config' in user_data[1])


class TestSeedData(unittest.TestCase):

    def setUp(self):
        super(TestSeedData, self).setUp()
        fixtures.setup_tests_config(self)
        self.conf = config.VmStack('foo')
        self.vm = vms.VM(self.conf)
        self.conf.set('vm.name', 'foo')
        self.conf.set('vm.class', 'lxd')
        self.conf.set('vm.release', 'focal')
        fixtures.override_logging(self)

    def test_create_meta_data(self):
        self.vm.create_meta_data()
        self.assertTrue(os.path.exists(self.vm._meta_data_path))

    def test_create_user_data(self):
        self.vm.create_user_data()
        self.assertTrue(os.path.exists(self.vm._user_data_path))


class TestKeyGen(unittest.TestCase):

    @features.requires(features.ssh_feature)
    def setUp(self):
        super(TestKeyGen, self).setUp()
        fixtures.isolate_from_disk(self)
        self.conf = config.VmStack(None)
        self.vm = vms.VM(self.conf)
        self.config_dir = os.path.join(self.uniq_dir, 'config')

    def load_config(self, more):
        content = '''\
vm.config_dir={config_dir}
vm.name=foo
'''.format(config_dir=self.config_dir)
        self.conf.store._load_from_string(content + more)

    def wrap_ssh_keygen(self):
        self.keygen_called = False

        orig_keygen = ssh.keygen

        def wrapped_key_gen(*args, **kwargs):
            self.keygen_called = True
            return orig_keygen(*args, **kwargs)

        fixtures.patch(self, ssh, 'keygen', wrapped_key_gen)

    def test_do_nothing_if_key_exists(self):
        self.load_config('ssh.server_keys=rsa')
        path = self.conf.get('ssh.server_keys')[0]
        self.wrap_ssh_keygen()
        with open(path, 'w') as f:
            f.write('something')
        self.vm.ssh_keygen()
        self.assertFalse(self.keygen_called)

    def test_recreate_key_if_force(self):
        self.load_config('ssh.server_keys=rsa')
        self.wrap_ssh_keygen()
        path = self.conf.get('ssh.server_keys')[0]
        with open(path, 'w') as f:
            f.write('something')
        self.vm.ssh_keygen(force=True)
        self.assertTrue(self.keygen_called)

    def test_create_key_if_it_doesnt_exist(self):
        self.load_config('ssh.server_keys=rsa')
        self.wrap_ssh_keygen()
        self.vm.ssh_keygen()
        self.assertTrue(self.keygen_called)
        self.assertTrue(os.path.exists('rsa'))
        self.assertTrue(os.path.exists('rsa.pub'))

    def assertKeygen(self, ssh_type, upper_type=None):
        if upper_type is None:
            upper_type = ssh_type.upper()
        self.load_config('ssh.server_keys={vm.config_dir}/' + ssh_type)
        self.vm.ssh_keygen()
        private_path = 'config/{}'.format(ssh_type)
        self.assertTrue(os.path.exists(private_path))
        public_path = 'config/{}.pub'.format(ssh_type)
        self.assertTrue(os.path.exists(public_path))

    def test_dsa(self):
        self.assertKeygen('dsa')

    def test_rsa(self):
        self.assertKeygen('rsa')

    def test_ecdsa(self):
        self.assertKeygen('ecdsa', 'EC')


class TestSSH(unittest.TestCase):

    def setUp(self):
        super(TestSSH, self).setUp()
        fixtures.isolate_from_disk(self)
        self.conf = config.VmStack(None)
        self.vm = vms.VM(self.conf)
        self.config_dir = os.path.join(self.uniq_dir, 'config')

    def load_config(self, more):
        content = '''\
vm.name=foo
ssh.options= # Disable the default value to make tests simpler
'''
        self.conf.store._load_from_string(content + more)

    def test_defaults(self):
        self.load_config('vm.ip = localhost')
        ssh_cmd = self.vm.get_ssh_command('whoami')
        self.assertEqual(
            ['ssh', self.vm.conf.get('vm.user') + '@localhost', 'whoami'],
            ssh_cmd)

    def test_ssh_user(self):
        self.load_config('vm.user = foo\nvm.ip = localhost')
        ssh_cmd = self.vm.get_ssh_command('whoami')
        self.assertEqual(['ssh', 'foo@localhost', 'whoami'], ssh_cmd)


class TestAptUpdate(unittest.TestCase):

    def setUp(self):
        super(TestAptUpdate, self).setUp()
        fixtures.isolate_from_disk(self)
        self.conf = config.VmStack(None)
        self.config_dir = os.path.join(self.uniq_dir, 'config')

        self.conf.store._load_from_string('''vm.update = true''')
        self.vm = vms.VM(self.conf)

    def test_apt_get_update_retries(self):
        self.vm.conf.set('apt.update.timeouts', '0.1, 0.1')
        self.nb_calls = 0

        def failing_update(command):
            self.nb_calls += 1
            if self.nb_calls > 1:
                return 0, 'stdout success', 'stderr success'
            else:
                # Fake a failed apt-get update
                raise errors.CommandError(['boo!'], 1, '', 'I failed')

        self.vm.do_apt_get = failing_update
        self.vm.apt_get_update()
        self.assertEqual(2, self.nb_calls)

    def test_apt_get_update_fails(self):
        self.vm.conf.set('apt.update.timeouts', '0.1, 0.1')
        self.nb_calls = 0

        def failing_update(command):
            self.nb_calls += 1
            raise errors.CommandError(['boo!'], 1, '', 'I failed')

        self.vm.do_apt_get = failing_update
        with self.assertRaises(errors.ByovError) as cm:
            self.vm.apt_get_update()
        self.assertEqual('apt-get update never succeeded',
                         str(cm.exception))
        self.assertEqual(2, self.nb_calls)


class TestWaitForIp(unittest.TestCase):

    def setUp(self):
        super(TestWaitForIp, self).setUp()
        fixtures.isolate_from_disk(self)
        self.conf = config.VmStack(None)
        self.vm = vms.VM(self.conf)

    def set_timeouts(self, value):
        # Use lxd.setup_ip.timeouts to get a TimeoutsOption
        self.conf.set('lxd.setup_ip.timeouts', value)
        self.vm.setup_ip_timeouts = 'lxd.setup_ip.timeouts'

    def test_wait_for_ip_fails(self):
        # Force a 0 timeout so the instance never get an IP
        self.set_timeouts('0, 0, 0')
        with self.assertRaises(errors.ByovError) as cm:
            self.vm.wait_for_ip()
        msg = '{} never received an IP'.format(self.conf.get('vm.name'))
        self.assertEqual(msg, str(cm.exception))

    def test_wait_for_ip_succeed(self):
        self.set_timeouts('0, 0, 1')

        def an_ip():
            return '127.0.0.1'
        self.vm.discover_ip = an_ip
        self.vm.wait_for_ip()
        self.assertEqual('127.0.0.1', self.vm.conf.get('vm.ip'))


# FIXME: parametrized by distribution -- vila 2017-12-06
# FIXME: Because launchpad.login is an ugly duck... -- vila 2016-11-07
# FIXME: Because gitlab.login is an ugly duck... -- vila 2018-07-25
@features.requires(features.bzr_feature)
class TestHashSetup(unittest.TestCase):

    def setUp(self):
        super(TestHashSetup, self).setUp()
        fixtures.isolate_from_disk(self)

    def prepare_vm(self, name='foo'):
        conf = config.VmStack(name)
        conf.set('vm.name', name)
        conf.set('vm.class', 'lxd')
        vm = vms.VM(conf)
        return vm

    def test_defaults(self):
        vm = self.prepare_vm()
        # This should not fail nor produce any output
        vm.hash_setup()

    def test_empty_options(self):
        vm = self.prepare_vm()
        vm.conf.set('vm.setup.digest.options', '')
        self.assertIsNot(None, vm.hash_setup())

    def test_change_name(self):
        foo = self.prepare_vm()
        foo_digest = foo.hash_setup()
        bar = self.prepare_vm('bar')
        bar_digest = bar.hash_setup()
        self.assertNotEqual(foo_digest, bar_digest)

    def test_unknown_file(self):
        vm = self.prepare_vm()
        vm.conf.set('vm.root_script', 'I-dont-exist')
        # This should not fail nor produce any output
        vm.hash_setup()

    def test_file_content_change(self):
        vm = self.prepare_vm()
        vm.conf.set('vm.root_script', 'script')
        with open('script', 'w') as f:
            f.write('do something')
        something = vm.hash_setup()
        with open('script', 'w') as f:
            f.write('do something else')
        something_else = vm.hash_setup()
        self.assertNotEqual(something, something_else)

    def test_file_name_change(self):
        vm = self.prepare_vm()
        vm.conf.set('vm.root_script', 'script')
        with open('script', 'w') as f:
            f.write('do something')
        something = vm.hash_setup()
        os.rename('script', 'tagada')
        vm.conf.set('vm.root_script', 'tagada')
        something_else = vm.hash_setup()
        self.assertEqual(something, something_else)

    def _test_package_file_content_change(self, oname):
        vm = self.prepare_vm()
        vm.conf.set(oname, '@packages')
        with open('packages', 'w') as f:
            f.write('a-package\nanother_package\n')
        something = vm.hash_setup()
        with open('packages', 'w') as f:
            f.write('another-package\n')
        something_else = vm.hash_setup()
        self.assertNotEqual(something, something_else)

    def test_apt_package_file_content_change(self):
        self._test_package_file_content_change('vm.packages')

    def test_pip_package_file_content_change(self):
        self._test_package_file_content_change('pip.packages')



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import unittest


from byov import config
from byov.vms import scaleway
from byov.tests import (
    features,
    fixtures,
)


@features.requires(features.scaleway_creds)
class TestIps(unittest.TestCase):

    def setUp(self):
        super(TestIps, self).setUp()
        self.conf = config.VmStack(None)
        self.client = scaleway.ComputeClient(
            self.conf, self.conf.get('scaleway.compute.url'),
            self.conf.get('scaleway.compute.timeouts'))

    def test_list_ips(self):
        # Short of having a dedicated account for tests, little can be assumed
        # about the list content. So this is mostly a smoke test.
        ips = self.client.ips_list()
        if ips:
            # When IPs exist, they all belong to the same organization
            foreigns = []
            organization = self.conf.get('scaleway.access_key')
            for ip in ips:
                if not ip['organization'] == organization:
                    foreigns.append(ip)
            self.assertEqual([], foreigns)

    def test_create_ip(self):
        old_ips = self.client.ips_list()
        nb_ips_before = len(old_ips)
        new_ip = self.client.create_ip()
        self.addCleanup(self.client.delete_ip, new_ip['id'])
        self.assertEqual(self.conf.get('scaleway.access_key'),
                         new_ip['organization'])
        new_ips = self.client.ips_list()
        nb_ips_after = len(new_ips)
        self.assertTrue(nb_ips_after > nb_ips_before,
                        'No IP was created: {} vs {}'.format(
                            old_ips, new_ips))

    def test_delete_ip(self):
        old_ips = self.client.ips_list()
        nb_ips_before = len(old_ips)
        new_ip = self.client.create_ip()
        self.assertEqual(self.conf.get('scaleway.access_key'),
                         new_ip['organization'])
        resp = self.client.delete_ip(new_ip['id'])
        self.assertTrue(resp.ok)
        self.assertEqual(204, resp.status_code)
        new_ips = self.client.ips_list()
        nb_ips_after = len(new_ips)
        self.assertTrue(nb_ips_after == nb_ips_before,
                        'No IP was deleted: {} vs {}'.format(
                            old_ips, new_ips))


@features.requires(features.scaleway_creds)
class TestVmWithIp(unittest.TestCase):

    def setUp(self):
        super(TestVmWithIp, self).setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        fixtures.override_logging(self)
        # Create a shared config
        self.conf = config.VmStack(self.vm_name)
        self.conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = scaleway
vm.distribution = ubuntu
vm.release = bionic
vm.architecture = amd64
'''.format(vm_name=self.vm_name))
        self.conf.set('scaleway.flavor', 'START1-XS')
        self.conf.set('scaleway.image',
                      '{vm.distribution}/{vm.release}/{vm.architecture}-xs')
        features.test_requires(self, features.ScalewayImage(self.conf))
        self.client = scaleway.ComputeClient(
            self.conf, self.conf.get('scaleway.compute.url'),
            self.conf.get('scaleway.compute.timeouts'))
        self.conf.store.save()
        self.conf.store.unload()

    def test_vm_with_ip(self):
        new_ip = self.client.create_ip()
        self.addCleanup(self.client.delete_ip, new_ip['id'])
        conf = config.VmStack(self.vm_name)
        conf.set('scaleway.public_ip', new_ip['address'])
        conf.set('vm.update', 'false')
        vm = scaleway.Scaleway(conf)
        self.addCleanup(vm.teardown, True)
        vm.setup()
        self.assertEqual(new_ip['address'], vm.conf.get('vm.ip'))



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import hashlib
import json
import logging
import sys
import time

try:
    import urlparse  # python2
except ImportError:
    from urllib import parse as urlparse

import byov
from byov import (
    errors,
    vms,
)

# FIXME: These are dependencies that need to remain optional so the scaleway
# backend requires them but byov doesn't -- vila 2018-01-14
import requests

logger = logging.getLogger(__name__)


# scaleway uses a different nomenclature. Probably closer to linux than ubuntu
# or lxd.
scaleway_architectures = dict(
    armhf='arm',
    amd64='x86_64',
    arm64='arm64',
)


class ScalewayComputeException(errors.ByovError):
    pass


class Client(object):
    """A client for the scaleway API.

    This is a simple wrapper around requests.Session so we inherit all good
    bits while providing a simple point for tests to override when needed.
    """

    user_agent = 'byov-{} python {}'.format(byov.version(),
                                            sys.version.split()[0])

    session_class = requests.Session

    def __init__(self, conf, root_url, timeouts):
        self.conf = conf
        self.root_url = root_url
        self.timeouts = timeouts
        self.session = self.session_class()

    def request(self, method, url, params=None, headers=None, **kwargs):
        """Overriding base class to handle the root url."""
        # Note that url may be absolute in which case 'root_url' is ignored by
        # urljoin.
        sent_headers = {'User-Agent': self.user_agent,
                        'X-Auth-Token': self.conf.get('scaleway.token'),
                        'Content-Type': 'application/json'}
        if headers is not None:
            sent_headers.update(headers)
        final_url = urlparse.urljoin(self.root_url,
                                     self.conf.expand_options(url))
        response = self.session.request(
            method, final_url, headers=sent_headers, params=params, **kwargs)
        return response

    def get(self, url, **kwargs):
        return self.request('GET', url, **kwargs)

    def post(self, url, data=None, **kwargs):
        if data is not None:
            data = json.dumps(data)
        return self.request('POST', url, data=data, **kwargs)

    def patch(self, url, **kwargs):
        return self.request('PATCH', url, **kwargs)

    def delete(self, url, **kwargs):
        return self.request('DELETE', url, **kwargs)

    def close(self):
        self.session.close()

    def retry(self, func, url, *args, **kwargs):
        req_path = '{} {}'.format(func.__name__.upper(), url)
        no_404_retry = kwargs.pop('no_404_retry', False)
        tmouts = self.timeouts
        for attempt, sleep in enumerate(tmouts, start=1):
            try:
                response = None
                if attempt > 1:
                    logger.debug('Re-trying {} {}/{}'.format(
                        req_path, attempt, tmouts.retries))
                response = func(url, *args, **kwargs)
            except requests.ConnectionError:
                # Most common transient failure: the API server is unreachable
                # (server, network or client may be the cause).
                msg = 'Connection error for {}, will sleep for {:.2f} seconds'
                logger.warning(msg.format(req_path, sleep))
            except Exception:
                # All other exceptions are raised
                logger.error('{} failed'.format(req_path), exc_info=True)
                raise ScalewayComputeException('{req} failed', req=req_path)
            # If the request succeeded return the response (also for the 404
            # special case as instructed).
            if (response is not None and
                (response.ok or (response.status_code == 404 and
                                 no_404_retry))):
                return response
            if response is not None and not response.ok:
                if response.status_code == 429:
                    msg = ('Rate limit reached for {},'
                           ' will sleep for {:.2f} seconds')
                    # This happens rarely but breaks badly if not caught. elmo
                    # recommended a 30 seconds nap in that case.
                    sleep += 30
                    logger.warning(msg.format(req_path, sleep))
                else:
                    # All other errors are raised
                    msg = '{req} failed {status}: {resp}'
                    # Nice spot to debug the API usage
                    logger.error(msg.format(req=req_path,
                                            status=response.status_code,
                                            resp=response),
                                 exc_info=True)
                    # FIXME: Debug, should not land -- vila 2018-01-18
                    sys.stderr.write(
                        msg.format(req=req_path,
                                   status=response.status_code,
                                   resp=response) + '\n')
                    raise ScalewayComputeException(msg, req=req_path,
                                                   status=response.status_code,
                                                   resp=response)
            # Take a nap before retrying
            logger.debug('Sleeping {:.2f} seconds for {} {}/{}'.format(
                sleep, req_path, attempt, tmouts.retries))
            time.sleep(sleep)
        # Raise if we didn't succeed at all
        raise ScalewayComputeException(
            "Failed to '{}' after {} retries".format(req_path, attempt))


class ComputeClient(object):
    """Client for the scaleway compute API."""

    http_class = Client

    def __init__(self, conf, root_url, timeouts):
        self.conf = conf
        self.http = self.http_class(conf, root_url, timeouts)

    def create_server(self, image_id, ip):
        data = dict(name=self.conf.get('vm.name'),
                    organization=self.conf.get('scaleway.access_key'),
                    image=image_id,
                    commercial_type=self.conf.get('scaleway.flavor'),
                    dynamic_ip_required=True,
                    volumes={})
        if ip:
            data['dynamic_ip_required'] = False
            data['public_ip'] = ip['id']
        else:
            data['dynamic_ip_required'] = True
        if self.conf.get('scaleway.boot.local'):
            # MISSINGTESTS: It works and is tested by being the only way to get
            # zfs working on some configurations :-/ No idea (yet) on how to
            # detect it on a running system -- vila 2018-10-15
            data['boot_type'] = 'local'
        response = self.http.retry(self.http.post, 'servers', data=data)
        return response.json()['server']

    def get_server_details(self):
        server_id = self.conf.get('scaleway.server_id')
        if server_id is None:
            # We can't ask about a server that doesn't exist yet
            return None
        response = self.http.retry(
            self.http.get, 'servers/{}'.format(server_id), no_404_retry=True)
        if response.status_code == 404:
            return dict(state='UNKNOWN')
        return response.json()['server']

    def set_user_data(self, key, data):
        return self.http.retry(
            self.http.patch,
            'servers/{}/user_data/{}'.format(
                self.conf.get('scaleway.server_id'), key),
            headers={'Content-Type': 'text/plain'},
            data=data)

    def start_server(self):
        self.action('poweron')

    def stop_server(self):
        # FIXME: Attempting to stop a stopped server 400's.
        self.action('poweroff')

    def terminate_server(self):
        # FIXME: Attempting to stop a stopped server 400's.
        self.action('terminate')

    def action(self, action):
        server_id = self.conf.get('scaleway.server_id')
        if server_id is None:
            # We can't act on a server that doesn't exist
            # FIXME: Building a 404 seems overkill -- vila 2018-01-15
            return None
        response = self.http.retry(
            self.http.post, 'servers/{}/action'.format(server_id),
            data=dict(action=action))
        return response

    def delete_server(self):
        server_id = self.conf.get('scaleway.server_id')
        if server_id is None:
            # We can't delete a server that doesn't exist
            # FIXME: Building a 404 seems overkill -- vila 2018-01-15
            return None
        response = self.http.retry(
            self.http.delete, 'servers/{}'.format(server_id))
        return response

    def delete_volume(self, volume_id):
        # FIXME: no_404_retry ? -- vila 2018-01-17
        response = self.http.retry(
            self.http.delete, 'volumes/{}'.format(volume_id))
        return response

    def create_snapshot(self, volume_id):
        response = self.http.retry(
            self.http.post, 'snapshots',
            data=dict(name=self.conf.get('vm.name'),
                      organization=self.conf.get('scaleway.access_key'),
                      volume_id=volume_id))
        return response.json()['snapshot']

    def delete_snapshot(self, snapshot_id):
        # FIXME: no_404_retry ? -- vila 2018-01-23
        response = self.http.retry(
            self.http.delete, 'snapshots/{}'.format(snapshot_id))
        return response

    def images_list(self):
        response = self.http.retry(self.http.get, 'images')
        while True:
            for image in response.json()['images']:
                yield image
            next_link = response.links.get('next', None)
            if next_link is None:
                break
            response = self.http.retry(self.http.get, next_link['url'])

    def create_image(self, name, arch, snapshot_id):
        response = self.http.retry(
            self.http.post, 'images',
            data=dict(name=name,
                      arch=arch,
                      organization=self.conf.get('scaleway.access_key'),
                      root_volume=snapshot_id))
        return response.json()['image']

    def delete_image(self, image_id):
        # FIXME: no_404_retry ? -- vila 2018-01-23
        response = self.http.retry(
            self.http.delete, 'images/{}'.format(image_id))
        return response

    def ips_list(self):
        response = self.http.retry(self.http.get, 'ips')
        return response.json()['ips']

    def create_ip(self):
        response = self.http.retry(
            self.http.post, 'ips',
            data=dict(organization=self.conf.get('scaleway.access_key')))
        return response.json()['ip']

    def delete_ip(self, ip_id):
        # FIXME: no_404_retry ? -- vila 2018-01-23
        response = self.http.retry(
            self.http.delete, 'ips/{}'.format(ip_id))
        return response


class Scaleway(vms.VM):

    setup_ip_timeouts = 'scaleway.setup_ip.timeouts'
    setup_ssh_timeouts = 'scaleway.setup_ssh.timeouts'

    compute_class = ComputeClient

    def __init__(self, conf):
        super(Scaleway, self).__init__(conf)
        self.compute = self.compute_class(
            self.conf,
            self.conf.get('scaleway.compute.url'),
            self.conf.get('scaleway.compute.timeouts'))
        self.server_id = None

    def state(self):
        # FIXME: The possible values need to be discovered and mapped
        # -- vila 20180-01-14
        # scw inspect can fail IRL just before the server starts and after it
        # shuts down (and can't be used while it's off :-/)
        # Seen IRL:
        # "state": "starting",
        # "state_detail": "provisioning node",
        # "state": "starting",
        # "state_detail": "booting kernel",
        # "state": "running",
        # "state_detail": "booted",
        # "state": "stopping",
        # "state_detail": "rebooting",
        # "state": "stopping",
        # "state_detail": "stopping",
        # curl to the rescue:
        # "state": "stopped"
        # "state_detail": ""
        server = self.compute.get_server_details()
        if server is None:
            return 'UNKNOWN'
        else:
            return server['state'].upper()

    def find_scaleway_image(self, image_name, arch):
        logger.debug('Searching for image {}...'.format(image_name))
        # First match wins
        existing_images = self.compute.images_list()
        for existing in existing_images:
            # FIXME: For convenience/debug, the image id can be used to select
            # a specific image. Beware though that this would allow an attacker
            # to use more recent images similarly named and a different id. In
            # other words: more work is needed to better select images. The
            # hidden assumption here is that the anti-chronological order is
            # convenient ;) -- # vila 2018-05-30
            if ((image_name == existing['name'] or
                 image_name == existing['id']) and arch == existing['arch']):
                logger.debug('Found image {}'.format(existing['id']))
                # More recent images come first. Picking the first match should
                # be the Right Thing.
                return existing
        raise ScalewayComputeException(
            'Image "{}" cannot be found'.format(image_name))

    def find_scaleway_reserved_ip(self, ip):
        logger.debug('Searching for ip {}...'.format(ip))
        # First match wins
        ips = self.compute.ips_list()
        for existing in ips:
            if ip == existing['address'] and existing['server'] is None:
                return existing
        raise ScalewayComputeException(
            'IP "{}" cannot be found'.format(ip))

    def create(self):
        logger.debug('Creating scaleway server {}'.format(
            self.conf.get('vm.name')))
        arch = scaleway_architectures[self.conf.get('vm.architecture')]
        image = self.find_scaleway_image(self.conf.get('scaleway.image'), arch)
        wanted_ip = self.conf.get('scaleway.public_ip')
        if wanted_ip:
            ip = self.find_scaleway_reserved_ip(wanted_ip)
        else:
            ip = None
        server = self.compute.create_server(image['id'], ip)
        self.econf.set('scaleway.server_id', server['id'])

    def set_cloud_init_config(self):
        self.create_user_data()
        logger.debug('Configuring scaleway server {}'.format(
            self.conf.get('vm.name')))
        self.compute.set_user_data('cloud-init', self.ci_user_data.dump())

    def setup(self):
        # We'll save the econf.store often so that when the setup fails, there
        # is something to debug.
        logger.info('Setting up scaleway server {}...'.format(
            self.conf.get('vm.name')))
        self.create()
        self.set_cloud_init_config()
        self.econf.set('scaleway.region_name',
                       self.conf.get('scaleway.region_name'))
        self.econf.store.save_changes()
        self.start()
        self.econf.store.save_changes()
        # MISSINGTEST: with bootstrap=True (though building the images count as
        # a manual but mandatory test to get anywhere -- vila 2018-01-24
        if not self.conf.get('scaleway.image.bootstrap'):
            self.wait_for_cloud_init()
        self.setup_over_ssh()

    def hash_setup(self):
        hasher = hashlib.md5()
        # Seed with the digest from the base class to capture common options
        hasher.update(super(Scaleway, self).hash_setup().encode('utf8'))
        options = self.conf.get('scaleway.setup.digest.options')
        if options:
            for opt in options:
                value = self.conf.get(opt, convert=False)
                self.hash_value(hasher, value)
        return hasher.hexdigest()

    def discover_ip(self):
        ip = None
        server = self.compute.get_server_details()
        state = server['state'].upper()
        public_ip = server.get('public_ip')
        if public_ip:
            ip = public_ip.get('address')
        if state != 'RUNNING' or not ip:
            raise ScalewayComputeException(
                'scaleway server {name} has not provided an IP yet: {state}',
                name=self.conf.get('vm.name'), state=state)
        return ip

    def wait_for_server_reaching(self, expected_states, how_long):
        logger.debug('Waiting for the server {} to reach {} state...'.format(
            self.conf.get('vm.name'), expected_states))
        # An amd64 (but this seems to be the same for all arches) server is in
        # the 'starting' state including 'allocating node' (up to a min),
        # 'provisioning node' (up to a min), 'booting kernel' (10s),
        # 'kernel-started' (5s), 'booted'. It then reaches the 'running' state.
        server = self.compute.get_server_details()
        timeout_limit = time.time() + how_long
        while (time.time() < timeout_limit and
               server['state'].upper() not in expected_states):
            logger.debug(
                'Server state: {state}/{state_detail}'.format(**server))
            time.sleep(5)
            server = self.compute.get_server_details()
        if server['state'].upper() not in expected_states:
            msg = ('Server {id} never reach state {expected_states}'
                   ' (last status: {state}/{state_detail})'.format(
                       expected_states=expected_states, **server))
            raise ScalewayComputeException(msg)
        return server

    def wait_for_cloud_init(self):
        self._wait_for_cloud_init('scaleway.cloud_init.timeouts')

    def start(self):
        super(Scaleway, self).start()
        logger.info('Starting scaleway server {}...'.format(
            self.conf.get('vm.name')))
        self.econf.store.save_changes()  # Recovery checkpoint
        how_long = self.conf.get('scaleway.poweron_timeout')
        timeout_limit = time.time() + how_long
        server = self.compute.get_server_details()
        attempts = 0
        while time.time() < timeout_limit:
            attempts += 1
            self.compute.start_server()
            server = self.wait_for_server_reaching(
                ['RUNNING', 'STOPPED'], how_long)
            if server['state'].upper() == 'STOPPED':
                # It may happen that the 'allocating node' step ends without
                # starting the server (it's just stopped: allocation failed
                # silently). Try again.
                logger.warning(
                    'Server {} allocation failed silently, re-trying'.format(
                        self.conf.get('vm.name')))
                continue
            else:
                break
        if server['state'].upper() != 'RUNNING':
            msg = ('Server {id} never reach state RUNNING'
                   ' (last status: {state}/{state_detail})'
                   ' after {attempts} attempts'.format(attempts=attempts,
                                                       **server))
            raise ScalewayComputeException(msg)

        self.wait_for_ip()
        self.econf.store.save_changes()  # Recovery checkpoint
        self.wait_for_ssh()
        self.save_existing_config()

    def stop(self):
        logger.info('Stopping scaleway server {}...'.format(
            self.conf.get('vm.name')))
        self.compute.stop_server()
        self.wait_for_server_reaching(
            ['STOPPED'], self.conf.get('scaleway.poweroff_timeout'))

        # FIXME: Hmm, volumes are archived by default but when tearing down
        # it's worth skipping the volume archive step as it takes a significant
        # time compared to 'stop --terminate'. In any case, it's probably a
        # good idea to keep track of the existing volumes in self.econf to be
        # able to clean them up by default (the only workflow that needs to
        # keep a boot volume is publish' AFAICS) -- vila 2018-01-14

    def publish(self):
        image_name = self.conf.get('vm.published_as')
        logger.info('Publishing scaleway image {} from {}...'.format(
            image_name, self.conf.get('vm.name')))
        server = self.compute.get_server_details()
        vol0 = server['volumes']['0']
        snapshot = self.compute.create_snapshot(vol0['id'])
        arch = scaleway_architectures[self.conf.get('vm.architecture')]
        image = self.compute.create_image(image_name, arch, snapshot['id'])
        return image

    def unpublish(self):
        image_name = self.conf.get('vm.published_as')
        logger.info('Un-publishing scaleway image {} from {}...'.format(
            image_name, self.conf.get('vm.name')))
        arch = scaleway_architectures[self.conf.get('vm.architecture')]
        image = self.find_scaleway_image(image_name, arch)
        snapshot_id = image['root_volume']['id']
        self.compute.delete_image(image['id'])
        self.compute.delete_snapshot(snapshot_id)

    def teardown(self, force=False):
        logger.info('Tearing down {} scaleway server...'.format(
            self.conf.get('vm.name')))
        if force and self.state() == 'RUNNING':
            self.compute.terminate_server()
            self.wait_for_server_reaching(
                ['UNKNOWN'], self.conf.get('scaleway.terminate_timeout'))
        else:
            # Get the server description before deletion
            server = self.compute.get_server_details()
            self.compute.delete_server()
            # Clean up the volumes
            # MISSINGTEST: race with server deletion ? what if server deletion
            # somehow deleted some volumes ? -- vila 2018-01-17
            if server['volumes']:
                for volume in server['volumes'].values():
                    self.compute.delete_volume(volume['id'])
        super(Scaleway, self).teardown()
        # FIXME: Beware of volume leaks -- vila 2018-01-14



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import logging
import os
import sys
import unittest

from byoc import errors
from byot import (
    assertions,
    scenarii,
)
from byov import (
    config,
    options,
)
from byov.tests import (
    features,
    fixtures,
)
from byov.vms import (
    lxd,
    subprocesses,
)

load_tests = scenarii.load_tests_with_scenarios


class TestVmMatcher(unittest.TestCase):

    def setUp(self):
        super(TestVmMatcher, self).setUp()
        fixtures.isolate_from_disk(self)
        self.store = config.VmStore('foo.conf')
        self.matcher = config.VmMatcher(self.store, 'quux')

    def test_empty_section_always_matches(self):
        self.store._load_from_string('foo=bar')
        matching = list(self.matcher.get_sections())
        self.assertEqual(1, len(matching))

    def test_specific_before_generic(self):
        self.store._load_from_string('foo=bar\n[quux]\nfoo=baz')
        matching = list(self.matcher.get_sections())
        self.assertEqual(2, len(matching))
        # First matching section is for quux
        self.assertEqual(self.store, matching[0][0])
        base_section = matching[0][1]
        self.assertEqual('quux', base_section.id)
        # Second matching section is the no-name one
        self.assertEqual(self.store, matching[0][0])
        no_name_section = matching[1][1]
        self.assertIs(None, no_name_section.id)


class TestVmStackOrdering(unittest.TestCase):

    def setUp(self):
        super(TestVmStackOrdering, self).setUp()
        fixtures.isolate_from_disk(self)
        self.conf = config.VmStack('foo')

    def assertValue(self, value, name):
        conf = config.VmStack('foo')
        self.assertEqual(value, conf.get(name))

    def test_default_in_empty_stack(self):
        self.assertValue('2048', 'vm.ram_size')

    def test_system_overrides_internal(self):
        with open(os.path.join(self.etc_dir,
                               config.config_file_basename()), 'w') as f:
            f.write('vm.ram_size = 42')
        self.assertValue('42', 'vm.ram_size')

    def test_user_overrides_system(self):
        with open(os.path.join(self.etc_dir,
                               config.config_file_basename()), 'w') as f:
            f.write('vm.ram_size = 42')
        os.makedirs(config.user_config_dir())
        with open(os.path.join(config.user_config_dir(),
                               config.config_file_basename()), 'w') as f:
            f.write('vm.ram_size = 84')
        self.assertValue('84', 'vm.ram_size')

    def test_local_overrides_user(self):
        os.makedirs(config.user_config_dir())
        with open(os.path.join(config.user_config_dir(),
                               config.config_file_basename()), 'w') as f:
            f.write('vm.ram_size = 84')
        with open('byov.conf', 'w') as f:
            f.write('vm.ram_size = 168')
        self.assertValue('168', 'vm.ram_size')

    def test_local_overrides_local_dir(self):
        os.makedirs('byov.conf.d')
        with open(os.path.join('byov.conf.d',
                               'foo.conf'), 'w') as f:
            f.write('vm.ram_size = 136')
        with open('byov.conf', 'w') as f:
            f.write('vm.ram_size = 168')
        self.assertValue('168', 'vm.ram_size')
        self.assertValue('168', 'vm.ram_size')

    def test_user_overrides_user_dir(self):
        os.makedirs(os.path.join(config.user_config_dir(), 'conf.d'))
        with open(os.path.join(config.user_config_dir(), 'conf.d',
                               'foo.conf'), 'w') as f:
            f.write('vm.ram_size = 136')
        with open(os.path.join(config.user_config_dir(),
                               config.config_file_basename()), 'w') as f:
            f.write('vm.ram_size = 84')
        self.assertValue('84', 'vm.ram_size')

    def test_system_overrides_system_dir(self):
        os.makedirs(os.path.join(config.system_config_dir(), 'conf.d'))
        with open(os.path.join(config.system_config_dir(), 'conf.d',
                               'foo.conf'), 'w') as f:
            f.write('vm.ram_size = 136')
        with open(os.path.join(config.system_config_dir(),
                               config.config_file_basename()), 'w') as f:
            f.write('vm.ram_size = 84')
        self.assertValue('84', 'vm.ram_size')

    def test_only_dotconf_are_seen(self):
        conf_d = os.path.join(config.user_config_dir(), 'conf.d')
        os.makedirs(conf_d)
        with open(os.path.join(conf_d, 'foo.conf'), 'w') as f:
            f.write('vm.ram_size = 42')
        with open(os.path.join(conf_d, 'foo'), 'w') as f:
            f.write('vm.ram_size = 84')
        self.assertValue('42', 'vm.ram_size')

    def test_dotconf_are_sorted(self):
        conf_d = os.path.join(config.user_config_dir(), 'conf.d')
        os.makedirs(conf_d)
        with open(os.path.join(conf_d, '1.conf'), 'w') as f:
            f.write('vm.ram_size = 1')
        with open(os.path.join(conf_d, '2.conf'), 'w') as f:
            f.write('vm.ram_size = 2')
        self.assertValue('1', 'vm.ram_size')


class TestVmStack(unittest.TestCase):
    """Test config option values."""

    def setUp(self):
        super(TestVmStack, self).setUp()
        fixtures.isolate_from_disk(self)
        self.conf = config.VmStack('foo')
        self.conf.store._load_from_string('''
vm.release=focal
vm.architecture=amd64
''')

    def assertValue(self, expected_value, option):
        self.assertEqual(expected_value, self.conf.get(option))

    def test_cloud_image_url(self):
        self.assertValue('http://cloud-images.ubuntu.com/focal/current/'
                         'focal-server-cloudimg-amd64.img',
                         'qemu.download.url')

    def test_apt_proxy_set(self):
        proxy = 'apt_proxy: http://example.org:4321'
        self.conf.set('apt.proxy', proxy)
        self.conf.store.save_changes()
        self.assertEqual(proxy, self.conf.expand_options('{apt.proxy}'))

    def test_vms_dir_with_user_expansion(self):
        vms_dir = '~/vms'
        self.conf.set('vm.vms_dir', vms_dir)
        self.conf.store.save_changes()
        self.assertValue(os.path.join(self.home_dir, 'vms'), 'vm.vms_dir')

    def test_vm_config_dir_with_user_expansion(self):
        config_dir = '~/config'
        self.conf.set('vm.config_dir', config_dir)
        self.conf.store.save_changes()
        self.assertValue(os.path.join(self.home_dir, 'config'),
                         'vm.config_dir')

    def test_lxd_invalid_nesting(self):
        self.conf.set('lxd.nesting', 'true')
        with self.assertRaises(errors.OptionValueError) as cm:
            self.conf.get('lxd.nesting')
        self.assertEqual('lxd.nesting: Value "true" is not valid:'
                         ' invalid literal for int() with base 10:'
                         " 'true'.",
                         str(cm.exception))


class TestPathOptionIsolated(unittest.TestCase):

    def setUp(self):
        super(TestPathOptionIsolated, self).setUp()
        fixtures.isolate_from_disk(self)

    def assertConverted(self, expected, value):
        option = options.PathOption('foo', help_string='A path.')
        self.assertEqual(expected, option.convert_from_unicode(None, value))

    def test_absolute_path(self):
        self.assertConverted('/test/path', '/test/path')

    def test_home_path_with_expansion(self):
        self.assertConverted(self.home_dir, '~')

    def test_path_in_home_with_expansion(self):
        self.assertConverted(os.path.join(self.home_dir, 'test/path'),
                             '~/test/path')


class TestLxdUserMountsOption(unittest.TestCase):

    def setUp(self):
        super(TestLxdUserMountsOption, self).setUp()
        fixtures.isolate_from_disk(self)

    def assertConverted(self, expected, value):
        option = options.LxdUserMountsOption('foo', help_string='A path.')
        self.assertEqual(expected, option.convert_from_unicode(None, value))

    def test_empty(self):
        self.assertConverted([], '')

    def test_odd(self):
        with self.assertRaises(errors.OptionValueError) as cm:
            self.assertConverted(None, 'tagada')
        self.assertEqual('foo', cm.exception.name)

    def test_vm_path_not_absolute(self):
        with self.assertRaises(errors.OptionValueError) as cm:
            self.assertConverted(None, 'here, there')
        self.assertEqual('foo', cm.exception.name)

    def test_absolute_paths(self):
        self.assertConverted([('/host/foo', '/vm/bar')], '/host/foo:/vm/bar')

    def test_tilde_paths(self):
        self.assertConverted([('{}/foo'.format(self.home_dir), '/bar')],
                             '~/foo:/bar')

    def test_relative_paths(self):
        self.assertConverted([('{}/foo'.format(self.uniq_dir), '/bar'),
                              (self.uniq_dir, '/quux')],
                             'foo:/bar, .:/quux')


class TestLoggingLevelOption(unittest.TestCase):

    def test_lower(self):
        self.assertEqual(logging.CRITICAL,
                         options.level_from_store('critical'))
        self.assertEqual(logging.ERROR, options.level_from_store('error'))
        self.assertEqual(logging.DEBUG, options.level_from_store('debug'))
        self.assertEqual(logging.NOTSET, options.level_from_store('notset'))

    def test_upper(self):
        self.assertEqual(logging.CRITICAL,
                         options.level_from_store('CRITICAL'))
        self.assertEqual(logging.ERROR, options.level_from_store('ERROR'))
        self.assertEqual(logging.DEBUG, options.level_from_store('DEBUG'))
        self.assertEqual(logging.NOTSET, options.level_from_store('NOTSET'))


@features.requires(features.bzr_feature)
class TestLaunchpadOptions(unittest.TestCase):

    def setUp(self):
        fixtures.isolate_from_disk(self)
        fixtures.override_env(self, 'BZR_HOME', self.uniq_dir)

    def test_defaults_to_logname(self):
        fixtures.override_env(self, 'LOGNAME', 'foo')
        self.assertEqual('foo', options.default_lp_login())

    # FIXME: testing bzr lp-login requires either a launchpad test server or a
    # real user with a registered ssh key, punting for now after manual tests
    # -- vila 2016-08-30
    def xtest_lp_login_set(self):
        subprocesses.run(['bzr', 'lp-login', 'foo'])
        self.assertEqual('foo', options.default_lp_login())


class TestPackageListOption(unittest.TestCase):

    def setUp(self):
        super(TestPackageListOption, self).setUp()
        fixtures.isolate_from_disk(self)

    def assertConverted(self, expected, value):
        option = options.PackageListOption('foo',
                                           help_string='A package list.')
        self.assertEqual(expected, option.convert_from_unicode(None, value))

    def test_empty(self):
        self.assertConverted(None, None)

    def test_regular_packages(self):
        self.assertConverted(['a', 'b'], 'a,b')

    def test_existing_file(self):
        with open('packages', 'w') as f:
            f.write('a\nb\n')
        self.assertConverted(['a', 'b'], '@packages')

    def test_comments(self):
        with open('packages', 'w') as f:
            f.write('#a\nb # EOL\nc\n')
        self.assertConverted(['b', 'c'], '@packages')

    def test_non_existing_file(self):
        with self.assertRaises(errors.OptionValueError) as cm:
            self.assertConverted(None, '@I-dont-exist')
        self.assertEqual('foo', cm.exception.name)


class TestVmUserOptions(unittest.TestCase):

    def setUp(self):
        super(TestVmUserOptions, self).setUp()
        fixtures.isolate_from_disk(self)

    def test_defaults(self):
        conf = config.VmStack('whatever')
        self.assertEqual('ubuntu', conf.get('vm.user'))
        self.assertEqual('/home/ubuntu', conf.get('vm.user.home'))
        self.assertEqual('/bin/bash', conf.get('vm.user.shell'))


class TestVmClass(unittest.TestCase):

    def setUp(self):
        super(TestVmClass, self).setUp()
        fixtures.isolate_from_disk(self)

    def test_class_mandatory(self):
        conf = config.VmStack('I-dont-exist')
        with self.assertRaises(errors.OptionMandatoryValueError):
            conf.get('vm.class')

    def test_bogus(self):
        conf = config.VmStack('I-dont-exist')
        conf.store._load_from_string('''vm.class=bogus''')
        with self.assertRaises(errors.OptionMandatoryValueError):
            conf.get('vm.class')


class TestAllVmClasses(unittest.TestCase):

    scenarios = [(k, dict(kls=k)) for k in options.vm_class_registry.keys()]

    def setUp(self):
        super(TestAllVmClasses, self).setUp()
        fixtures.isolate_from_disk(self)

    def test_backend(self):
        conf = config.VmStack('I-dont-exist')
        conf.store._load_from_string('vm.class={}'.format(self.kls))
        self.assertIs(options.vm_class_registry.get(self.kls),
                      conf.get('vm.class'))


class TestStartingNameMatcher(unittest.TestCase):

    def setUp(self):
        super(TestStartingNameMatcher, self).setUp()
        fixtures.isolate_from_disk(self)
        # Any simple store is good enough
        self.store = config.VmStore('foo.conf')

    def assertSectionIDs(self, expected, name, content):
        self.store._load_from_string(content)
        matcher = config.StartingNameMatcher(self.store, name)
        sections = list(matcher.get_sections())
        assertions.assertLength(self, len(expected), sections)
        self.assertEqual(expected, [section.id for _, section in sections])
        return sections

    def test_none_with_sections(self):
        self.assertSectionIDs([], None, '''\
[foo]
[bar]
''')

    def test_empty(self):
        self.assertSectionIDs([], 'whatever', '')

    def test_matching_paths(self):
        self.assertSectionIDs(['foo-bar', 'foo'],
                              'foo-bar-baz', '''\
[foo]
[foo-bar]
''')

    def test_no_name_section_included_when_present(self):
        # Note that other tests will cover the case where the no-name section
        # is empty and as such, not included.
        self.assertSectionIDs(['foo-bar', 'foo', None],
                              'foo-bar-baz', '''\
option = defined so the no-name section exists
[foo]
[foo-bar]
''')

    def test_order_reversed(self):
        self.assertSectionIDs(['foo-bar', 'foo'], 'foo-bar-baz', '''\
[foo]
[foo-bar]
''')

    def test_unrelated_section_excluded(self):
        self.assertSectionIDs(['foo-bar', 'foo'], 'foo-bar-baz', '''\
[foo]
[foo-qux]
[foo-bar]
''')


class TestVmHostOrdering(unittest.TestCase):

    def setUp(self):
        super(TestVmHostOrdering, self).setUp()
        fixtures.isolate_from_disk(self)

    def assertSectionIDs(self, expected, conf):
        actual = [section.id for (store, section) in conf.iter_sections()]
        self.assertEqual(expected, actual)

    def test_respect_host_order_in_single_file(self):
        with open('byov.conf', 'w') as f:
            f.write('''\
[foo-bar-baz]
[foo-qux]
[foo-b]
[foo]
''')
        conf = config.VmStack('foo-bar-baz')
        self.assertSectionIDs([None, 'foo-bar-baz', 'foo-b', 'foo'], conf)

    def test_respect_host_order_across_files(self):
        with open('byov.conf', 'w') as f:
            f.write('''\
[foo]
''')
        os.mkdir('byov.conf.d')
        with open('byov.conf.d/foo.conf', 'w') as f:
            f.write('''\
[foo-b]
''')
        conf = config.VmStack('foo-bar-baz')
        self.assertSectionIDs([None, 'foo-b', 'foo'], conf)


class TestVmName(unittest.TestCase):

    def setUp(self):
        super(TestVmName, self).setUp()
        fixtures.isolate_from_disk(self)

    def save_config(self, content):
        os.makedirs(config.user_config_dir())
        user_config_path = os.path.join(config.user_config_dir(),
                                        config.config_file_basename())
        with open(user_config_path, 'w')as f:
            f.write(content)

    def test_name_starts_with_section_name(self):
        self.save_config('''\
[foo]
vm.class = lxd
''')
        conf = config.VmStack('foo1')
        conf.cmdline_store.from_cmdline(['vm.name=foo1'])
        self.assertEqual('foo1', conf.get('vm.name'))
        self.assertIs(lxd.Lxd, conf.get('vm.class'))


class TestExistingVmStack(unittest.TestCase):

    def setUp(self):
        super(TestExistingVmStack, self).setUp()
        fixtures.isolate_from_disk(self)

    def test_empty(self):
        conf = config.ExistingVmStack('foo')
        self.assertEqual([], list(conf.iter_sections()))

    def test_existing_vm(self):
        econf = config.ExistingVmStack('foo')
        econf.set('vm.name', 'foo')
        econf.set('vm.class', 'lxd')  # We need a valid class
        econf.set('vm.ip', 'localhost')
        econf.store.save_changes()
        conf = config.VmStack('foo')
        self.assertEqual('foo', conf.get('vm.name'))
        self.assertEqual(lxd.Lxd, conf.get('vm.class'))
        self.assertEqual('localhost', conf.get('vm.ip'))

    def test_sibling_vms_dont_leak(self):
        foo_econf = config.ExistingVmStack('foo')
        foo_econf.set('vm.name', 'foo')
        foo_econf.set('vm.class', 'lxd')  # We need a valid class
        foo_econf.set('blah', 'blah')
        foo_econf.store.save_changes()
        foo2_econf = config.ExistingVmStack('foo-2')
        foo2_econf.set('vm.name', 'foo-2')
        foo2_econf.set('vm.class', 'lxd')
        foo2_econf.store.save_changes()

        conf = config.VmStack('foo-2')
        self.assertEqual('foo-2', conf.get('vm.name'))
        self.assertEqual(lxd.Lxd, conf.get('vm.class'))
        self.assertIsNone(conf.get('blah'))


class TestUserByov(unittest.TestCase):

    # The byov imported file change the value
    changed_by = None

    def setUp(self):
        super(TestUserByov, self).setUp()
        fixtures.isolate_from_disk(self)
        if sys.version_info < (3,):
            self.skipTest('no support for user byos in python2')

    def create_import(self, directory, value):
        """Create a byov.py file in the directory setting changed to True."""
        os.makedirs(directory)
        with open(os.path.join(directory, 'byov.py'), 'w') as f:
            f.write('import {}\n'. format(self.__class__.__module__))
            f.write('{}.{}.changed = "{}"\n'.format(
                self.__class__.__module__, self.__class__.__name__, value))

    def test_local(self):
        self.create_import('byov.conf.d', 'local')
        config.import_user_byovs()
        self.assertEqual('local', self.changed)

    def test_user(self):
        self.create_import(os.path.join(config.user_config_dir(), 'conf.d'),
                           'user')
        config.import_user_byovs()
        self.assertEqual('user', self.changed)

    def test_system(self):
        self.create_import(os.path.join(config.system_config_dir(), 'conf.d'),
                           'system')
        config.import_user_byovs()
        self.assertEqual('system', self.changed)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2015, 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import errno
import hashlib
import logging
import os
import time


from novaclient import exceptions
from novaclient import client

# v2 is available from wily and the python3 client from xenial
try:
    from novaclient import v2 as nova_api
    nova_api.version = '2'
except ImportError:
    from novaclient import v1_1 as nova_api
    nova_api.version = '1'

from byov import (
    errors,
    timeouts,
    vms,
)


logger = logging.getLogger(__name__)


def byov_image_name(domain, series, architecture):
    """Returns an image name.

    The images are uploaded to glance for specific needs.

    :param domain: 'cloudimg' or 'britney'.

    :param series: The ubuntu series (precise, trusty, etc).

    :param architecture: The processor architecture ('amd64', i386, etc).
    """
    if domain not in ('cloudimg', 'britney'):
        raise ValueError('Invalid image domain')
    return 'byov/{}/{}-{}.img'.format(domain, series, architecture)


def get_os_nova_client(conf):
    os_nova_client = client.Client(
        nova_api.version,
        conf.get('nova.username'), conf.get('nova.password'),
        conf.get('nova.tenant_name'),
        conf.get('nova.auth_url'),
        region_name=conf.get('nova.region_name'),
        service_type='compute')
    return os_nova_client


class NovaServerException(errors.ByovError):
    pass


class NovaClient(object):
    """A nova client re-trying requests on known transient failures."""

    def __init__(self, conf, **kwargs):
        self.first_wait = kwargs.pop('first_wait', 30)
        self.wait_up_to = kwargs.pop('wait_up_to', 600)
        self.retries = kwargs.pop('retries', 8)
        self.nova = get_os_nova_client(conf)

    def retry(self, func, *args, **kwargs):
        no_404_retry = kwargs.pop('no_404_retry', False)
        sleeps = timeouts.ExponentialBackoff(
            self.first_wait, self.wait_up_to, self.retries)
        for attempt, sleep in enumerate(sleeps, start=1):
            try:
                if attempt > 1:
                    logger.warning('Re-trying {} {}/{}'.format(
                        func.__name__, attempt, self.retries))
                return func(*args, **kwargs)
            except client.requests.ConnectionError:
                # Most common transient failure: the API server is unreachable
                msg = 'Connection error for {}, will sleep for {:.2f} seconds'
                logger.warning(msg.format(func.__name__, sleep))
            except (exceptions.OverLimit, exceptions.RateLimit):
                msg = ('Rate limit reached for {},'
                       ' will sleep for {:.2f} seconds')
                # This happens rarely but breaks badly if not caught. elmo
                # recommended a 30 seconds nap in that case.
                sleep += 30
                logger.warning(msg.format(func.__name__, sleep))
            except exceptions.ClientException as e:
                if no_404_retry and e.http_status == 404:
                    raise
                msg = '{} failed will sleep for {:.2f} seconds'
                logger.warning(msg.format(func.__name__, sleep))
            except Exception:
                # All other exceptions are raised
                logger.error('{} failed'.format(func.__name__), exc_info=True)
                raise NovaServerException('{name} failed', name=func.__name__)
            # Take a nap before retrying
            logger.info('Sleeping {:.2f} seconds for {} {}/{}'.format(
                sleep, func.__name__, attempt, self.retries))
            time.sleep(sleep)
        # Raise if we didn't succeed at all
        raise NovaServerException("Failed to '{name}' after {attempt} retries",
                                  name=func.__name__, attempt=attempt)

    def flavors_list(self):
        return self.retry(self.nova.flavors.list)

    def images_list(self):
        return self.retry(self.nova.images.list)

    def create_server(self, name, flavor, image, user_data, nics,
                      availability_zone):
        return self.retry(self.nova.servers.create, name=name,
                          flavor=flavor, image=image, userdata=user_data,
                          nics=nics, availability_zone=availability_zone)

    def delete_server(self, server_id):
        # FIXME: 404 shouldn't be retried, if it's not there anymore, there is
        # nothing to delete. -- vila 2015-07-16
        return self.retry(self.nova.servers.delete, server_id)

    def start_server(self, instance):
        return self.retry(instance.start)

    def stop_server(self, instance):
        return self.retry(instance.stop)

    def create_floating_ip(self):
        return self.retry(self.nova.floating_ips.create)

    def delete_floating_ip(self, floating_ip):
        return self.retry(self.nova.floating_ips.delete, floating_ip)

    def add_floating_ip(self, instance, floating_ip):
        return self.retry(instance.add_floating_ip, floating_ip)

    def get_server_details(self, server_id):
        return self.retry(self.nova.servers.get, server_id,
                          no_404_retry=True)

    def get_server_console(self, server, length=None):
        return self.retry(server.get_console_output, length)


class NovaServer(vms.VM):

    nova_client_class = NovaClient
    setup_ip_timeouts = 'nova.setup_ip.timeouts'
    setup_ssh_timeouts = 'nova.setup_ssh.timeouts'

    def __init__(self, conf):
        super(NovaServer, self).__init__(conf)
        self.instance = None
        self.floating_ip = None
        self.nova = self.build_nova_client()
        self.econf.set('vm.final_message', 'testbed setup completed.')

    # MISSINGTEST
    def state(self):
        try:
            with open(self.nova_id_path()) as f:
                nova_id = f.read().strip()
        except IOError as e:
            # python2 does not provide FileNotFoundError
            if e.errno == errno.ENOENT:
                # Unknown interface
                return 'UNKNOWN'
        try:
            self.instance = self.nova.get_server_details(nova_id)
        except exceptions.NotFound:
            return 'UNKNOWN'
        # The instance may remain in the DELETED state for some time.
        nova_states = dict(BUILD='STARTING',
                           ACTIVE='RUNNING',
                           SHUTOFF='STOPPED',
                           DELETED='UNKNOWN')
        return nova_states[self.instance.status]

    def build_nova_client(self):
        nova_client = self.nova_client_class(self.conf)
        return nova_client

    def find_flavor(self):
        flavors = self.conf.get('nova.flavors')
        if not flavors:
            raise NovaServerException('nova.flavors must be set')
        logger.debug('Searching for a valid flavor in {}'.format(
            ' '.join(flavors)))
        existing_flavors = self.nova.flavors_list()
        for flavor in flavors:
            for existing in existing_flavors:
                if flavor == existing.name:
                    return existing
        raise NovaServerException(
            'None of [{flavors}] can be found', flavors=','.join(flavors))

    def find_nova_image(self):
        image_name = self.conf.get('nova.image')
        logger.debug('Searching for image {}...'.format(image_name))
        existing_images = self.nova.images_list()
        for existing in existing_images:
            if image_name == existing.name:
                return existing
        raise NovaServerException(
            'Image "{name}" cannot be found', name=image_name)

    def find_nics(self):
        net_id = self.conf.get('nova.net_id')
        if net_id:
            return [{'net-id': self.conf.get('nova.net_id')}]
        return None

    # FIXME: This should save the console whether or not the setup fails
    # -- vila 2015-08-26
    def setup(self):
        logger.info('Setting up {} nova server...'.format(
            self.conf.get('vm.name')))
        flavor = self.find_flavor()
        image = self.find_nova_image()
        nics = self.find_nics()
        self.create_user_data()
        with open(self._user_data_path) as f:
            user_data = f.read()
        logger.debug('Creating {} nova server...'.format(
            self.conf.get('vm.name')))
        self.instance = self.nova.create_server(
            name=self.conf.get('vm.name'), flavor=flavor, image=image,
            user_data=user_data, nics=nics,
            # FIXME: We probably want at least a vm.az_name option. And get
            # that option from higher levels too -- vila 2014-10-13
            availability_zone=None)
        self.create_nova_id_file(self.instance.id)
        self.wait_for_active_instance()
# FIXME: We want a vm.create_floating_ip option ? -- vila 2015-08-24
#        if unit_config.is_hpcloud(self.conf.get('os.auth_url')):
#            self.floating_ip = self.nova.create_floating_ip()
#            self.nova.add_floating_ip(self.instance, self.floating_ip)
        self.wait_for_ip()
        self.wait_for_cloud_init()
        self.wait_for_ssh()
        self.setup_over_ssh()

    def hash_setup(self):
        hasher = hashlib.md5()
        # Seed with the digest from the base class to capture common options
        hasher.update(super(NovaServer, self).hash_setup().encode('utf8'))
        options = self.conf.get('nova.setup.digest.options')
        if options:
            for opt in options:
                value = self.conf.get(opt, convert=False)
                self.hash_value(hasher, value)
        return hasher.hexdigest()

    def update_instance(self, nova_id=None):
        if nova_id is None:
            nova_id = self.instance.id
        try:
            # Always query nova to get updated data about the instance
            self.instance = self.nova.get_server_details(nova_id)
            return True
        except Exception:
            logger.debug('Cannot update instance {}'.format(nova_id),
                         exc_info=True)
            # But catch exceptions if something goes wrong. Higher levels will
            # deal with the instance not replying.
            return False

    def wait_for_active_instance(self):
        logger.debug('Waiting for the instance to become active...')
        timeout_limit = time.time() + self.conf.get('nova.boot_timeout')
        while (time.time() < timeout_limit and
               self.instance.status not in ('ACTIVE', 'ERROR')):
            time.sleep(5)
            self.update_instance()
        if self.instance.status != 'ACTIVE':
            msg = 'Instance {instance} never came up (last status: {status})'
            raise NovaServerException(msg, instance=self.instance.id,
                                      status=self.instance.status)

    def nova_id_path(self):
        # FIXME: This should be in self.econf -- vila 2018-01-14
        return os.path.join(self.config_dir_path(), 'nova_id')

    def create_nova_id_file(self, nova_id):
        nova_id_path = self.nova_id_path()
        self.ensure_dir(self.config_dir_path())
        with open(nova_id_path, 'w') as f:
            f.write(nova_id + '\n')

    def discover_ip(self):
        networks = self.instance.networks.values()
        if not networks:
            raise NovaServerException('No network for {instance}',
                                      instance=self.instance.id)
        # The network name is arbitrary, can vary for different clouds
        # but there should be only one network so we get the first one
        # and avoid the need for a config option for the network name.
        # We take the last IP address so it's either the only one or
        # the floating one. In both cases that gives us a reachable IP.
        return networks[0][-1]

    def get_cloud_init_console(self, length=None):
        return self.nova.get_server_console(self.instance, length)

    # FIXME: ~Duplicated with lxc, refactor to extract discovering cloud-init
    # completion (success/failure may be refined later if at least ssh access
    # is provided. In this case, the cloud-init log files can be
    # acquired/analyzed more precisely. This will require refactoring
    # setup(). -- vila 2015-10-25
    def wait_for_cloud_init(self):
        logger.info('Waiting for cloud-init...')
        timeout_limit = (time.time() +
                         self.conf.get('nova.cloud_init_timeout'))
        final_message = self.conf.get('vm.final_message')
        while time.time() < timeout_limit:
            # A relatively cheap way to catch cloud-init completion is to watch
            # the console for the specific message we specified in user-data).
            # FIXME: or at least check that we don't miss when we sleep a
            # significant time between two calls (like on canonistack where we
            # can sleep for minute(s) -- vila 2015-07-17
            console = self.get_cloud_init_console(20)
            if final_message in console:
                # We're good to go
                logger.info(
                    'cloud-init completed for {}'.format(self.instance.id))
                return
            time.sleep(5)
        raise NovaServerException(
            'Instance {instance} never completed cloud-init',
            instance=self.instance.id)

    def start(self):
        super(NovaServer, self).start()
        logger.info('Starting {} nova server...'.format(
            self.conf.get('vm.name')))
        self.nova.start_server(self.instance)
        self.wait_for_ip()
        self.wait_for_ssh()
        self.save_existing_config()

    # MISSINGTEST
    def stop(self):
        logger.info('Stopping {} nova server...'.format(
            self.conf.get('vm.name')))
        self.nova.stop_server(self.instance)
        # FIXME: Should wait for the instance to be shut off -- vila 2016-07-01
        # With a wait parameter defaulting to True ? -- vila 2017-01-20

    def teardown(self, force=False):
        logger.info('Tearing down {} nova server...'.format(
            self.conf.get('vm.name')))
        if force and self.state() == 'RUNNING':
            self.stop()
        if self.instance is not None:
            logger.info('Deleting instance {}'.format(self.instance.id))
            self.nova.delete_server(self.instance.id)
            # FIXME: Should wait for the instance to disappear
            # -- vila 2016-07-01
            self.instance = None
            os.remove(self.nova_id_path())
        if self.floating_ip is not None:
            self.nova.delete_floating_ip(self.floating_ip)
            self.floating_ip = None
        super(NovaServer, self).teardown()



#!/usr/bin/env python3

# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014, 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import setuptools


import byov


class CleanCommand(setuptools.Command):
    """Custom clean command to tidy up the project root."""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./byov.egg-info')


setuptools.setup(
    name='byov',
    version='.'.join(str(c) for c in byov.__version__[0:3]),
    description=('Build Your Own Virtual machine.'),
    author='Vincent Ladeuil',
    author_email='v.ladeuil+lp@free.fr',
    url='https://launchpad.net/byov',
    license='GPLv3',
    install_requires=['byoc'],
    packages=['byov', 'byov.tests', 'byov.vms'],
    # FIXME: byovm is kept for backwards compatibility, remove it for 3.0
    # -- vila 2022-03-09
    entry_points=dict(console_scripts=['byov=byov.commands:run',
                                       'byovm=byov.commands:run']),
    cmdclass=dict(clean=CleanCommand),
)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import os
import subprocess
import sys

try:
    if sys.version_info < (3,):
        # novaclient doesn't support python3 (yet)
        from byov.vms import nova
        with_nova = True
    else:
        with_nova = False
except ImportError:
    with_nova = False

try:
    from byov.vms import ec2
    with_ec2 = True
except ImportError:
    with_ec2 = False

from byoc import errors
from byot import features
from byov import (
    config,
    subprocesses,
)
from byov.vms import (
    docker,
    scaleway,
)


class SshFeature(features.ExecutableFeature):

    def __init__(self):
        super(SshFeature, self).__init__('ssh')
        self.version = None

    def _probe(self):
        exists = super(SshFeature, self)._probe()
        if exists:
            try:
                proc = subprocess.Popen(['ssh', '-V'],
                                        stderr=subprocess.PIPE,
                                        stdout=subprocess.PIPE)
                out, err = proc.communicate()
            except OSError as e:
                if e.errno == errno.ENOENT:
                    # broken install
                    return False
                else:
                    raise
            self.version = err.decode()
        return exists


def requires_existing_vm(test, vm_name):
    """Skip a test if a known reference vm is not provided locally.

    :note: This may be revisited later when it becomes easier to create vms for
       test purposes.

    :note: This should be called after fixtures.setup_tests_config() has been
    called so the user-provided test config is installed.

    :param test: The TestCase requiring the vm.

    :param vm_name: The vm name in the config.

    """
    user_conf = config.VmStack(vm_name)
    # FIXME: This is suspicious and seem to explain why vm.name is set in test
    # config files -- vila 2018-07-29
    if user_conf.get('vm.name') != vm_name:
        test.skipTest('{} is not defined'.format(vm_name))
    kls = user_conf.get('vm.class')
    vm = kls(user_conf)
    if vm.state() == 'UNKNOWN':
        test.skipTest('{} has not been setup'.format(vm_name))


def requires_remote_lxd(test):
    """Skip a test if no configuration is found for lxd remote access."""
    remote_name = 'byov-tests-remote-lxd'
    conf = config.VmStack(remote_name)
    remote_lxd = conf.get('remote.name')
    if not remote_lxd:
        test.skipTest('{} is not defined'.format(remote_name))
    return remote_lxd


def requires_existing_bridge(test, br_name):
    """Skip a test if a bridge does not exists."""
    br_path = '/sys/class/net/{}/bridge'.format(br_name)
    # This doesn't verify the bridge is functional (how can this be achieved ?)
    # but can be used by any user (i.e. not root).
    if not os.path.exists(br_path):
        test.skipTest('{} is not defined'.format(br_path))


class NovaCompute(features.Feature):

    def _probe(self):
        if not with_nova:
            # novaclient doesn't support python3 (yet)
            return False
        client = self.get_client()
        if client is None:
            return False
        try:
            # can transiently fail with requests.exceptions.ConnectionError
            # (converted from MaxRetryError).
            client.authenticate()
        except nova.exceptions.ClientException:
            return False
        return True

    def get_client(self):
        # The byov-tests-nova vm is not required to exist, we'll get
        # credentials from the environment if they are defined there.
        test_vm_conf = config.VmStack('byov-tests-nova')
        try:
            return nova.get_os_nova_client(test_vm_conf)
        except errors.OptionMandatoryValueError:
            return None


class NovaCredentials(features.Feature):

    def _probe(self):
        if not with_nova:
            # novaclient doesn't support python3 (yet)
            return False
        try:
            config.VmStack('byov-tests-nova').get_nova_creds()
        except errors.OptionMandatoryValueError:
            return False
        return True

    def feature_name(self):
        return 'A valid set of nova credentials'


class Ec2Boto3(features.Feature):

    def _probe(self):
        if not with_ec2:
            return False
        client = self.get_client()
        if client is None:
            return False
        try:
            # t3.micro is available where t2.micro is not depending on regions
            client.instance_types(['t2.micro', 't3.micro'], True)
        except (ec2.Ec2ServerException, ec2.botocore.exceptions.ClientError):
            # When the credentials are invalid the botocore exception is
            # translated into an Ec2ServerException.
            return False
        return True

    def get_client(self):
        # FIXME: the byov-tests-ec2 does not exist -- vila 2021-12-03
        # The byov-tests-ec2 vm is not required to exist, we'll get
        # credentials from the environment if they are defined there.
        test_vm_conf = config.VmStack('byov-tests-ec2')
        try:
            return ec2.Ec2Client(test_vm_conf)
        except errors.OptionMandatoryValueError:
            return None

    def feature_name(self):
        return 'boto3 client version >= 1.21.0'


class Ec2Credentials(features.Feature):

    def _probe(self):
        if not with_ec2:
            return False
        try:
            # FIXME: the byov-tests-ec2 does not exist -- vila 2021-12-03
            config.VmStack('byov-tests-ec2').get_aws_creds()
        except errors.OptionMandatoryValueError:
            return False
        return True

    def feature_name(self):
        return 'A valid set of aws credentials'


class ScalewayCredentials(features.Feature):

    def _probe(self):
        try:
            config.VmStack('byov-tests-scaleway').get_scaleway_creds()
        except errors.OptionMandatoryValueError:
            return False
        return True

    def feature_name(self):
        return 'A valid set of scaleway credentials'


# FIXME: Strictly speaking, ScalewayImage should require
# ScalewayCredentials. In practice, it's easier to require a config object here
# which already knows the credentials but also provide the image name and
# architecture. -- vila 2018-10-14
class ScalewayImage(features.Feature):

    def __init__(self, conf):
        super(ScalewayImage, self).__init__()
        self.conf = conf
        self.image = self.conf.get('scaleway.image')
        self.architecture = scaleway.scaleway_architectures[
            self.conf.get('vm.architecture')]

    def _probe(self):
        vm = scaleway.Scaleway(self.conf)
        try:
            vm.find_scaleway_image(self.image, self.architecture)
        except scaleway.ScalewayComputeException:
            return False
        return True

    def feature_name(self):
        return 'Image {} at scaleway'.format(self.image)


class TestsConfig(features.Feature):
    """User provided tests configuration.

    Some tests require existing vms that the user needs to
    provide. 'byov.conf-tests' define the vms and should be under
    ~/.config/byov'.

    """

    def __init__(self):
        super(TestsConfig, self).__init__()
        self.user_path = os.path.join(config.user_config_dir(),
                                      'byov.conf-tests')
        self.more_paths = list(config.config_files_in(config.user_config_dir(),
                                                      suffix='.conf-tests'))

    def _probe(self):
        # We only test that the config files exists here, tests can make more
        # checks about the content (like ssh.key and the
        # byov-tests-lxd-debian vm)
        return os.path.exists(self.user_path)

    def feature_name(self):
        return 'User-provided configuration to setup test vms'


class LxdNesting(features.Feature):

    def __init__(self, level):
        super(LxdNesting, self).__init__()
        self.level = level

    def _probe(self):
        from byov.vms import lxd
        return lxd.check_nesting(self.level)

    def feature_name(self):
        return ('lxd configured to nest containers'
                ' up to {} level{}'.format(self.level,
                                           's' if self.level > 1 else ''))


# FIXME: This is broken by design. It should be a vm declared in the config
# files so tests that requires it can setup/teardown rather than leaving the
# container running -- vila 2022-01-19
class DockerRegistry(features.Feature):
    """On-demand docker image registry."""

    def _probe(self):
        conf = config.VmStack(None)
        info = docker.container_info(conf.get('docker.registry.name'))
        if info['state'] != 'RUNNING':
            docker_registry_command = conf.get('docker.registry.command')
            ret, out, err = subprocesses.run(docker_registry_command)
        return True

    def feature_name(self):
        return 'docker test image registry'


# There is always a single instance of a given feature (so it's probed only
# once) shared by all tests, they are all declared below to be easier to find.

# FIXME: Is it worth it to have wget as a soft dependency when downloads are
# easy enough to be handled at the http level ? Error handling and timeouts are
# not yet implemented around wget either and would also be simpler at the http
# level. -- vila 2019-11-08
bzr_feature = features.ExecutableFeature('bzr')
docker_client_feature = features.ExecutableFeature('docker')
ec2_boto3 = Ec2Boto3()
ec2_creds = Ec2Credentials()
git_feature = features.ExecutableFeature('git')
lxd_client_feature = features.ExecutableFeature('lxc')
lxd_nesting_1 = LxdNesting(1)
nova_compute = NovaCompute()
nova_creds = NovaCredentials()
qemu_img_feature = features.ExecutableFeature('qemu-img')
scaleway_creds = ScalewayCredentials()
ssh_feature = SshFeature()
tests_config = TestsConfig()
wget_feature = features.ExecutableFeature('wget')

# Useful shortcuts to export
requires = features.requires
test_requires = features.test_requires



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2015, 2016, 2017 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import random

# A very nice and concise summary about timeouts and retries and why using any
# network related resource requires them:
# https://docs.aws.amazon.com/general/latest/gr/api-retries.html


class ExponentialBackoff(object):
    """Provides wait times summing up to a limit.

    When an operation can fail transiently, it can be retried several times
    until it succeeds up to a specified limit. The returned values are the
    successive wait times and their total equals ``up_to``.

    :note: The total of the returned times will be ``up_to``. This mimics
        manual retries by a user re-trying randomly but giving up after a
        limit.

    :note: To simplify use by callers doing a loop around attempt/sleep, the
        last value is always zero so no time is wasted waiting after the last
        failed attempt.
    """

    random = random.random

    def __init__(self, first, up_to, retries):
        self.first = first
        self.up_to = up_to
        self.retries = retries

    def __repr__(self):
        return 'waiting {first} up to {up_to} in {retries} attempts'.format(
            **self.__dict__)

    def __iter__(self):
        attempts = 1
        cumulated = backoff = self.first
        if self.retries:
            # First wait is specified by the user
            yield self.first
        # Yield the time to wait between retries
        while attempts + 1 < self.retries:
            if cumulated + backoff > self.up_to:
                backoff = 0
            else:
                cumulated += backoff
            yield backoff
            backoff = self.first + (2 ** attempts) * self.random()
            attempts += 1
        if self.retries > 1:
            # We need at least two retries to reach ``up_to`` since ``first``
            # is specified by the user
            yield (self.up_to - cumulated)
        if self.retries:
            # Waiting after the last failure is useless
            yield 0.0



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import io
import sys
import yaml

import unittest

from byoc import errors as conf_errors

from byov import (
    config,
    errors,
    user_data,
)
from byov.tests import fixtures


class TestYaml(unittest.TestCase):

    def yaml_load(self, *args, **kwargs):
        return yaml.safe_load(*args, **kwargs)

    def yaml_dump(self, *args, **kwargs):
        return yaml.safe_dump(*args, **kwargs)

    def test_load_scalar(self):
        self.assertEqual(
            {'foo': 'bar'}, self.yaml_load(io.StringIO('{foo: bar}')))
        # Surprisingly the enclosing braces are not needed, probably a special
        # case for the highest level
        self.assertEqual({'foo': 'bar'}, self.yaml_load(
            io.StringIO('foo: bar')))

    def test_dump_scalar(self):
        if yaml.__version__ >= '5.1':
            expected = 'foo: bar\n'
        else:
            expected = '{foo: bar}\n'
        self.assertEqual(expected, self.yaml_dump(dict(foo='bar')))

    def test_load_list(self):
        self.assertEqual({'foo': ['a', 'b', 'c']},
                         # Single space indentation is enough
                         self.yaml_load(io.StringIO('''\
foo:
 - a
 - b
 - c
''')))

    def test_dump_list(self):
        # No more enclosing braces... yeah for consistency :-/
        if yaml.__version__ >= '5.1':
            expected = 'foo:\n- a\n- b\n- c\n'
        else:
            expected = 'foo: [a, b, c]\n'
        self.assertEqual(
            expected, self.yaml_dump(dict(foo=['a', 'b', 'c'])))

    def test_load_dict(self):
        self.assertEqual({'foo': {'bar': 'baz'}},
                         self.yaml_load(io.StringIO('{foo: {bar: baz}}')))
        multiple_lines = '''\
foo: {bar: multiple
  lines}
'''
        self.assertEqual(
            {'foo': {'bar': 'multiple lines'}},
            self.yaml_load(io.StringIO(multiple_lines)))


class TestCIUserData(unittest.TestCase):

    maxDiff = None

    def setUp(self):
        super(TestCIUserData, self).setUp()
        fixtures.setup_tests_config(self)
        self.conf = config.VmStack('foo')
        # Set some options to arbitrary but valid values
        self.conf.set('vm.name', 'foo')
        self.conf.set('vm.class', 'lxd')
        self.conf.set('vm.distribution', 'ubuntu')
        self.ci_data = user_data.CIUserData(self.conf)

    def test_empty_config(self):
        self.ci_data.populate()
        # Check default values
        self.assertIs(None, self.ci_data.root_hook)
        self.assertIs(None, self.ci_data.user_hook)
        cc = self.ci_data.cloud_config
        self.assertFalse('apt_update' in cc)
        self.assertFalse('apt_upgrade' in cc)
        self.assertEqual({'expire': True}, cc['chpasswd'])
        self.assertEqual('byov finished installing in ${uptime} seconds.',
                         cc['final_message'])
        self.assertEqual('True', cc['manage_etc_hosts'])
        self.assertFalse('fqdn' in cc)
        self.assertEqual('en_US.UTF-8', cc['locale'])

    def test_etc_manage_hosts_locahost(self):
        self.conf.set('vm.manage_etc_hosts', 'localhost')
        self.ci_data.populate()
        self.assertEqual('localhost',
                         self.ci_data.cloud_config['manage_etc_hosts'])

    def test_fqdn(self):
        fqdn = 'foo.example.com'
        self.conf.set('vm.fqdn', fqdn)
        self.ci_data.populate()
        self.assertEqual(fqdn, self.ci_data.cloud_config['fqdn'])

    def test_vm_user_default(self):
        self.conf.set('ssh.authorized_keys', '')
        self.ci_data.populate()
        self.assertEqual([{'name': 'ubuntu', 'homedir': '/home/ubuntu',
                           'shell': '/bin/bash',
                           'system': False,
                           'sudo': ['ALL=(ALL) NOPASSWD:ALL']}],
                         self.ci_data.cloud_config['users'])

    def test_specified_vm_user(self):
        self.conf.set('vm.user', 'tester')
        self.conf.set('vm.user.home', '/lamaison')
        self.conf.set('vm.user.shell', '/bin/bash')
        with open('test-key', 'w') as f:
            f.write('unused')
        self.conf.set('ssh.authorized_keys', 'test-key')
        self.conf.set('vm.user.sudo', 'ALL=(ALL) NOPASSWD:/bin/true')
        self.ci_data.populate()
        self.assertEqual([{'name': 'tester',
                           'shell': '/bin/bash',
                           'system': False,
                           'homedir': '/lamaison',
                           'ssh-authorized-keys': ['unused'],
                           'sudo': ['ALL=(ALL) NOPASSWD:/bin/true']}],
                         self.ci_data.cloud_config['users'])

    def test_vm_chpasswd_default(self):
        self.ci_data.populate()
        self.assertEqual({'expire': True},
                         self.ci_data.cloud_config['chpasswd'])

    def test_vm_chpasswd(self):
        self.conf.set('vm.chpasswd', 'root:s3cr3t, ubuntu:tagada')
        self.ci_data.populate()
        self.assertEqual(
            {'expire': False, 'list': ['root:s3cr3t', 'ubuntu:tagada']},
            self.ci_data.cloud_config['chpasswd'])

    def test_apt_proxy(self):
        self.conf.set('apt.proxy', 'tagada')
        self.ci_data.populate()
        self.assertEqual('tagada', self.ci_data.cloud_config['apt_proxy'])

    def test_locale(self):
        self.conf.set('vm.locale', 'C.UTF-8')
        self.ci_data.populate()
        self.assertEqual('C.UTF-8', self.ci_data.cloud_config['locale'])

    def test_final_message_precise(self):
        self.conf.set('vm.release', 'precise')
        self.ci_data.populate()
        self.assertEqual('byov finished installing in $UPTIME seconds.',
                         self.ci_data.cloud_config['final_message'])

    def test_final_message_explicit(self):
        self.conf.set('vm.final_message', 'hello there')
        self.ci_data.populate()
        self.assertEqual('hello there',
                         self.ci_data.cloud_config['final_message'])

    def test_update_true(self):
        self.conf.set('vm.update', 'True')
        self.ci_data.populate()
        cc = self.ci_data.cloud_config
        # We don't use cloud-init for that anymore, ensure we don't regress
        self.assertFalse('apt_update' in cc)
        self.assertFalse('apt_upgrade' in cc)
        self.assertFalse('package_update' in cc)
        self.assertFalse('package_upgrade' in cc)

    def test_apt_sources(self):
        self.conf.set('vm.release', 'RELEASE')
        self.conf.set('_archive_url', 'http://archive.ubuntu.com/ubuntu')
        self.conf.set('_ppa_url', 'https://u:p@ppa.lp.net/user/ppa/ubuntu')
        self.conf.set('apt.sources',
                      'deb {_archive_url} {vm.release} partner,'
                      ' deb {_archive_url} {vm.release} main,'
                      ' deb {_ppa_url} {vm.release} main|ABCDEF')
        self.ci_data.populate()
        self.assertEqual(
            [{'source':
              'deb http://archive.ubuntu.com/ubuntu RELEASE partner'},
             {'source': 'deb http://archive.ubuntu.com/ubuntu RELEASE main'},
             {'source':
              'deb https://u:p@ppa.lp.net/user/ppa/ubuntu RELEASE main',
              'keyid': 'ABCDEF'}],
            self.ci_data.cloud_config['apt_sources'])

    def create_file(self, path, content=None):
        if content is None:
            content = '{}\ncontent\n'.format(path)
        with open(path, 'wb') as f:
            if sys.version_info[0] < 3:
                f.write(content)
            else:
                f.write(bytes(content, 'utf8'))

    def test_good_ssh_keys(self):
        paths = ('rsa', 'rsa.pub', 'dsa', 'dsa.pub', 'ecdsa', 'ecdsa.pub')
        for path in paths:
            self.create_file(path)
        self.conf.set('ssh.server_keys', ','.join(paths))
        self.ci_data.populate()
        self.assertEqual({'dsa_private': 'dsa\ncontent\n',
                          'dsa_public': 'dsa.pub\ncontent\n',
                          'ecdsa_private': 'ecdsa\ncontent\n',
                          'ecdsa_public': 'ecdsa.pub\ncontent\n',
                          'rsa_private': 'rsa\ncontent\n',
                          'rsa_public': 'rsa.pub\ncontent\n'},
                         self.ci_data.cloud_config['ssh_keys'])

    def test_bad_type_ssh_keys(self):
        self.conf.set('ssh.server_keys', 'I-dont-exist')
        self.assertRaises(errors.ConfigValueError, self.ci_data.populate)

    def test_unknown_ssh_keys(self):
        self.conf.set('ssh.server_keys', 'rsa.pub')
        self.assertRaises(errors.ConfigPathNotFound, self.ci_data.populate)

    def test_unknown_ssh_authorized_keys(self):
        self.conf.set('ssh.authorized_keys', 'rsa.pub')
        self.assertRaises(errors.ConfigPathNotFound, self.ci_data.populate)

    def test_unknown_root_script(self):
        self.conf.set('vm.root_script', 'I-dont-exist')
        self.assertRaises(errors.ConfigPathNotFound, self.ci_data.populate)

    def test_unknown_user_script(self):
        self.conf.set('vm.user_script', 'I-dont-exist')
        self.assertRaises(errors.ConfigPathNotFound, self.ci_data.populate)

    def test_expansion_error_in_script(self):
        root_script_content = '''#!/bin/sh
echo Hello {I_dont_exist}
'''
        with open('root_script.sh', 'w') as f:
            f.write(root_script_content)
        self.conf.set('vm.root_script', 'root_script.sh')
        with self.assertRaises(conf_errors.ExpandingUnknownOption) as cm:
            self.ci_data.populate()
        self.assertEqual(root_script_content, cm.exception.string)

    def test_unknown_uploaded_scripts(self):
        self.conf.set('vm.uploaded_scripts', 'I-dont-exist')
        self.assertRaises(errors.ConfigPathNotFound, self.ci_data.populate)

    def test_root_script(self):
        with open('root_script.sh', 'w') as f:
            f.write('''#!/bin/sh
echo Hello {user}
''')
        self.conf.set('vm.root_script', 'root_script.sh')
        self.conf.set('user', 'root')
        self.ci_data.populate()
        # The additional newline after the script is expected
        self.assertEqual('''\
#!/bin/sh -e
cat >/tmp/root-byov_pre_install <<'EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN'
#!/bin/sh
echo Hello root

EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN
chown root:root /tmp/root-byov_pre_install
chmod 0700 /tmp/root-byov_pre_install
''', self.ci_data.root_hook)
        self.assertEqual(['/tmp/root-byov_pre_install'],
                         self.ci_data.cloud_config['bootcmd'])

    def test_user_script(self):
        with open('user_script.sh', 'w') as f:
            f.write('''#!/bin/sh
echo Hello {user}
''')
        self.conf.set('vm.user_script', 'user_script.sh')
        self.conf.set('user', 'testuser')
        self.ci_data.populate()
        # The additional newline after the script is expected
        self.assertEqual('''\
#!/bin/sh
echo Hello testuser
''', self.ci_data.user_hook)
        self.assertEqual(['/tmp/root-byov_post_install'],
                         self.ci_data.cloud_config['runcmd'])

    def test_uploaded_scripts(self):
        paths = ('foo', 'bar')
        for path in paths:
            self.create_file(path)
        self.conf.set('vm.uploaded_scripts', ','.join(paths))
        self.ci_data.populate()
        self.assertEqual('''\
#!/bin/sh -e
cat >~ubuntu/byov_uploads <<'EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN'
mkdir -p ~ubuntu/bin
cd ~ubuntu/bin
cat >foo <<'EOFfoo'
foo
content

EOFfoo
chmod 0755 foo
cat >bar <<'EOFbar'
bar
content

EOFbar
chmod 0755 bar
EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN
chown ubuntu:ubuntu ~ubuntu/byov_uploads
chmod 0700 ~ubuntu/byov_uploads
''',
                         self.ci_data.uploaded_scripts_hook)
        self.assertEqual([['su', '-l', '-c', '~ubuntu/byov_uploads',
                           'ubuntu']],
                         self.ci_data.cloud_config['runcmd'])
